# -*- coding: utf-8 -*-

# Copyright (c) 2019 Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""implements :class:`gpxity.backends.test.test_backends.TestBackends` for all backends."""

# pylint: disable=protected-access

import os
import time
import datetime
import random
import tempfile

from unittest import skipIf

from .basic import BasicTest, disabled
from .. import Directory, MMT, GPSIES, TrackMMT, Mailer, WPTrackserver, Openrunner
from ... import Track, Lifetrack, Backend, Account, DirectoryAccount
from ...util import remove_directory

# pylint: disable=attribute-defined-outside-init


class TestBackends(BasicTest):

    """Are the :literal:`supported_` attributes set correctly?."""

    def test_supported(self):
        """Check values in supported for all backends."""
        expect_unsupported = dict()
        expect_unsupported[Directory] = {
            'own_categories', 'write_add_keywords', 'write_remove_keywords',
            'write_category', 'write_description', 'write_public', 'write_title'}
        expect_unsupported[MMT] = {
            'rename', }
        expect_unsupported[GPSIES] = {
            'keywords', 'write_add_keywords', 'write_remove_keywords', 'rename'}
        expect_unsupported[TrackMMT] = {
            'scan', 'remove', 'rename',
            'write_title', 'write_description', 'write_public',
            'write_category', 'write_add_keywords',
            'write_remove_keywords'}
        expect_unsupported[Mailer] = {
            'own_categories', 'scan', 'remove', 'rename',
            'write_title', 'write_description', 'write_public',
            'write_category', 'write_add_keywords',
            'write_remove_keywords'}
        expect_unsupported[Openrunner] = {
            'write_title', 'write_description', 'write_public',
            'write_category', 'write_add_keywords', 'rename',
            'write_remove_keywords'}
        expect_unsupported[WPTrackserver] = {
            'own_categories',
            'write_add_keywords', 'write_remove_keywords', 'write_category',
            'write_description', 'write_public', 'write_title'}
        for cls in Backend.all_backend_classes():
            with self.tst_backend(cls):
                self.assertTrue(
                    cls.supported & expect_unsupported[cls] == set(),
                    '{}: supported & unsupported: {}'.format(
                        cls.__name__, cls.supported & expect_unsupported[cls]))
                self.assertEqual(
                    sorted(cls.supported | expect_unsupported[cls]),
                    sorted(cls.full_support),
                    '{}.supported is wrong'.format(cls.__name__))

    def test_all_backends(self):
        """Check if Backend.all_backend_classes works."""
        backends = Backend.all_backend_classes()
        expected = [Directory, GPSIES, MMT, Mailer, Openrunner, TrackMMT, WPTrackserver]
        expected = [x for x in expected if not x.is_disabled()]
        self.assertEqual(backends, expected)

    def test_account_none(self):
        """Test access with Account=None."""
        for cls in Backend.all_backend_classes():
            with self.tst_backend(cls):
                if cls is Directory:
                    backend = cls(DirectoryAccount())
                else:
                    backend = cls(Account())
                if cls in (GPSIES, MMT, Openrunner):
                    with self.assertRaises(backend.BackendException) as context:
                        backend.scan(now=True)
                    self.assertEqual(str(context.exception), '{} needs a username'.format(backend.url))

    def test_subscription(self):
        """Test backend.subscription."""
        for cls in Backend.all_backend_classes():
            with self.tst_backend(cls):
                with self.temp_backend(cls) as backend:
                    if cls in (MMT, Openrunner):
                        self.assertEqual(backend.subscription, 'free')
                    elif cls == TrackMMT:
                        self.assertEqual(backend.subscription, 'full')
                    else:
                        self.assertIsNone(backend.subscription)

    def test_save_empty(self):
        """Save empty track."""
        for cls in Backend.all_backend_classes(needs={'write'}):
            with self.tst_backend(cls):
                with self.temp_backend(cls) as backend:
                    track = Track()
                    if cls in (MMT, TrackMMT, GPSIES, Openrunner):
                        with self.assertRaises(cls.BackendException):
                            backend.add(track)
                    else:
                        self.assertIsNotNone(backend.add(track))

    @skipIf(*disabled(Directory))
    def test_directory_backend(self):
        """Manipulate backend."""
        track = self.create_test_track()
        with self.temp_backend(Directory) as directory1:
            with self.temp_backend(Directory) as directory2:
                saved = directory1.add(track)
                self.assertBackendLength(directory1, 1)
                self.assertEqual(saved.backend, directory1)
                directory1.add(track.clone())
                self.assertBackendLength(directory1, 2)
                directory2.add(track)
                self.assertBackendLength(directory2, 1)
                directory2.scan()
                self.assertBackendLength(directory2, 1)

    def test_duplicate_tracks(self):
        """What happens if we save the same track twice?."""
        for cls in Backend.all_backend_classes(needs={'remove', 'write'}):
            with self.tst_backend(cls):
                with self.temp_backend(cls) as backend:
                    track = self.create_test_track(cls)
                    backend.add(track)
                    self.assertBackendLength(backend, 1)
                    with self.assertRaises(ValueError):
                        backend.add(track)
                    self.assertBackendLength(backend, 1)
                    if cls is GPSIES:
                        # if the same track data is uploaded again, we get the same id_in_backend.
                        with self.assertRaises(ValueError):
                            backend.add(track.clone())
                        self.assertBackendLength(backend, 1)
                    else:
                        backend.add(track.clone())
                        self.assertBackendLength(backend, 2)

    def test_open_wrong_username(self):
        """Open backends with username missing in Account."""
        for cls in Backend.all_backend_classes(exclude=[Directory, Mailer, TrackMMT, Openrunner]):
            # Openrunner allows arbitrary user names
            with self.tst_backend(cls):
                with self.assertRaises(cls.BackendException):
                    self.setup_backend(cls, test_name='wrong_username').scan(now=True)

    def test_open_wrong_password(self):
        """Open backends with wrong password."""
        for cls in Backend.all_backend_classes(needs={'scan'}):
            with self.tst_backend(cls):
                if not issubclass(cls, Directory):
                    with self.assertRaises(cls.BackendException):
                        self.setup_backend(cls, test_name='wrong_password').scan(now=True)

    @skipIf(*disabled(Directory))
    def test_match(self):
        """test backend match function.

        Returns:
            None

        """
        def match_date(track) ->str:
            """match against a date.

            Returns:
                None if match else an error message
            """
            if track.first_time < datetime.datetime(year=2016, month=9, day=5):
                return 'time {} is before {}'.format(track.first_time, '2016-09-05')
            return None

        cls = Directory
        with self.tst_backend(cls):
            with self.temp_backend(cls, count=3) as backend:
                for idx, _ in enumerate(backend):
                    _.adjust_time(datetime.timedelta(hours=4 + idx))
                new_track = backend[0].clone()
                self.assertIsNotNone(match_date(new_track))
                self.assertBackendLength(backend, 3)
                backend.match = match_date
                self.assertBackendLength(backend, 1)
                with self.assertRaises(cls.NoMatch):
                    backend.add(new_track)
                self.assertBackendLength(backend, 1)
                orig_time = backend[0].first_time
                delta = datetime.timedelta(days=-5)
                with self.assertRaises(cls.NoMatch):
                    backend[0].adjust_time(delta)
                self.assertBackendLength(backend, 1)
                self.assertEqual(orig_time + delta, backend[0].first_time)

    def test_z9_create_backend(self):
        """Test creation of a backend."""
        for cls in Backend.all_backend_classes(needs={'remove'}):
            if not cls.test_is_expensive:
                with self.tst_backend(cls):
                    with self.temp_backend(cls, count=3) as backend:
                        self.assertBackendLength(backend, 3)
                        first_time = backend.get_time()
                        time.sleep(2)
                        second_time = backend.get_time()
                        total_seconds = (second_time - first_time).total_seconds()
                        self.assertTrue(1 < total_seconds < 8, 'Time difference should be {}, is {}-{}={}'.format(
                            2, second_time, first_time, second_time - first_time))

    def test_write_remoteattr(self):
        """If we change title, description, public, category in track, is the backend updated?."""
        for cls in Backend.all_backend_classes(needs={'remove', }):
            with self.tst_backend(cls):
                with self.temp_backend(cls, count=1) as backend:
                    track = backend[0]
                    first_title = track.title
                    first_description = track.description
                    track.title = 'A new title'
                    self.assertEqual(track.title, 'A new title')
                    track.description = 'A new description'
                    # make sure there is no cache in the way
                    backend2 = backend.clone()
                    track2 = backend2[0]
                    self.assertEqualTracks(track, track2)
                    self.assertNotEqual(first_title, track2.title)
                    self.assertNotEqual(first_description, track2.description)

    def test_write_category(self):
        """If we change category in track, is the backend updated?."""
        for cls in Backend.all_backend_classes(needs={'remove', }):
            with self.tst_backend(cls):
                test_category = cls.decode_category(cls.supported_categories[10])
                test_category2 = cls.decode_category(cls.supported_categories[15])
                with self.temp_backend(cls, count=1, category=test_category) as backend:
                    track = backend[0]
                    self.assertEqual(track.category, test_category)
                    track.category = test_category2
                    # make sure there is no cache in the way
                    backend2 = backend.clone()
                    track2 = backend2[0]
                    self.assertEqualTracks(track, track2, 'category should be {}'.format(test_category2))
                    self.assertEqual(track2.category, test_category2, 'category should be {}'.format(test_category2))

    def test_write_public(self):
        """If we change public in track, is the backend updated?."""
        for cls in Backend.all_backend_classes(needs={'remove', }):
            with self.tst_backend(cls):
                test_public = True
                test_public2 = False
                with self.temp_backend(cls, count=1, public=test_public) as backend:
                    track = backend[0]
                    orig_cat = track.category
                    self.assertEqual(track.public, test_public)
                    track.public = test_public2
                    self.assertEqual(track.category, orig_cat)
                    # make sure there is no cache in the way
                    backend2 = backend.clone()
                    track2 = backend2[0]
                    self.assertEqual(track2.category, orig_cat)
                    self.assertEqualTracks(track, track2)
                    self.assertEqual(track2.public, test_public2)

    def xtest_gpsies_bug(self):
        """We have this bug only sometimes: title, category or time will be wrong in track2.
        Workaround is in GPSIES._edit."""
        for _ in range(20):
            with self.temp_backend(GPSIES, count=1, category=GPSIES.supported_categories[3]) as backend:
                track = backend[0]
                track.title = 'A new title'
                track.description = 'A new description'
                track.category = backend.decode_category(backend.supported_categories[8])
                # make sure there is no cache in the way
                backend2 = backend.clone()
                track2 = backend2[0]
                self.assertEqualTracks(track, track2, with_category=True)

    def test_z2_keywords(self):
        """save and load keywords."""  # noqa

        kw_a = 'A'
        kw_b = 'Berlin'
        kw_c = 'CamelCase'
        kw_d = self.unicode_string2

        def minus(value):
            """

            Returns:
                value preceded with -

            """
            return '-' + value

        for cls in Backend.all_backend_classes(needs={'write', 'scan', 'keywords'}):
            with self.tst_backend(cls):
                with self.temp_backend(cls, clear_first=True, count=1) as backend:
                    # TODO: warum  noch clear_first?
                    track = backend[0]
                    track.keywords = list()
                    self.assertEqual(track.keywords, list())
                    track.keywords = ([kw_a, kw_b, kw_c])
                    track.change_keywords(minus(kw_b))
                    self.assertTrue(track is backend[0])
                    self.assertHasKeywords(track, (kw_a, kw_c))
                    with self.assertRaises(Exception):
                        track.change_keywords('Category:whatever')
                    track.change_keywords(kw_d)
                    self.assertHasKeywords(track, (kw_a, kw_c, kw_d))
                    backend2 = backend.clone()
                    track2 = backend2[track.id_in_backend]
                    self.assertTrue(track is backend[0])
                    track2.change_keywords(minus(kw_d))
                    self.assertHasKeywords(backend[0], (kw_a, kw_c, kw_d))
                    # change_keywords may have change id_in_backend in some backend classes, so reload track
                    backend.scan()
                    track = backend[0]
                    self.assertHasKeywords(track2, (kw_a, kw_c))
                    self.assertHasKeywords(track, (kw_a, kw_c))
                    self.assertTrue(track is backend[0])
                    backend.scan()
                    self.assertHasKeywords(track, (kw_a, kw_c))
                    track.change_keywords(minus(kw_a))
                    self.assertHasKeywords(track, [kw_c])
                    # track2.change_keywords(minus(kw_a))
                    track.change_keywords(minus(kw_c))
                    track.change_keywords(minus(kw_d))
                    backend.scan()
                    self.assertHasKeywords(backend[0], list())

    def test_z_unicode(self):
        """Can we up- and download unicode characters in all text attributes?."""
        tstdescr = 'DESCRIPTION with ' + self.unicode_string1 + ' and ' + self.unicode_string2
        for cls in Backend.all_backend_classes(needs={'remove'}):
            with self.tst_backend(cls):
                with self.temp_backend(cls, count=1) as backend:
                    backend2 = backend.clone()
                    track = backend[0]
                    self.assertIsNotNone(track.backend)
                    track.title = 'Title ' + self.unicode_string1
                    self.assertIsNotNone(track.backend)
                    self.assertEqual(track.backend, backend)
                    backend2.scan()  # because backend2 does not know about changes thru backend
                    track2 = backend2[0]
                    # track and track2 may not be identical. If the original track
                    # contains gpx xml data ignored by MMT, it will not be in track2.
                    self.assertEqual(track.title, track2.title)
                    track.description = tstdescr
                    self.assertEqual(track.description, tstdescr)
                    if cls is Directory:
                        self.assertTrackFileContains(track, tstdescr)
                    backend2.scan()
                    self.assertEqual(backend2[0].description, tstdescr)
                    backend2.detach()

    def test_change_points(self):
        """Can we change the points of a track?.

        For MMT this means re-uploading and removing the previous instance, so this

        is not always as trivial as it should be."""

    @skipIf(True, "enable manully if needed")
    def test_download_many_from_mmt(self):
        """Download many tracks."""
        many = 150
        with self.temp_backend(MMT, test_name='gpxstoragemany', count=many, clear_first=True) as backend:
            self.assertBackendLength(backend, many)

    def test_duplicate_title(self):
        """two tracks having the same title."""
        for cls in Backend.all_backend_classes(needs={'remove'}):
            with self.tst_backend(cls):
                with self.temp_backend(cls, count=2) as backend:
                    backend[0].title = 'TITLE'
                    backend[1].title = 'TITLE'

    @skipIf(*disabled(Directory))
    def test_private(self):
        """Up- and download private tracks."""
        with self.temp_backend(Directory) as local:
            # TODO: make cls outer loop and count for expensive cls
            track = self._get_track_from_test_file('test2')
            self.assertTrue(track.public)  # as defined in test2.gpx keywords
            track.public = False
            self.assertFalse(track.public)
            local.add(track)
            self.assertBackendLength(local, 1)
            for cls in Backend.all_backend_classes(needs={'remove'}):
                with self.tst_backend(cls):
                    with self.temp_backend(cls) as backend:
                        backend.merge(local)
                        for _ in backend:
                            self.assertFalse(_.public)
                        backend2 = backend.clone()
                        with self.temp_backend(Directory) as copy:
                            for _ in copy.merge(backend2):
                                self.logger.debug(_)
                            self.assertSameTracks(
                                local, copy,
                                with_last_time=cls not in (GPSIES, Openrunner),
                                with_category=False)

    @skipIf(*disabled(Directory))
    def test_merge_backends(self):
        """merge backends."""
        def dump(msg):
            for line in msg:
                self.logger.debug(line)
        org_source_len = 2
        assert org_source_len >= 2
        org_sink_len = 2
        assert org_sink_len >= 2
        with self.temp_backend(Directory, url='source', count=org_source_len) as source:
            with self.temp_backend(Directory, url='sink', count=org_sink_len) as sink:
                for _ in list(sink)[1:]:
                    _.adjust_time(datetime.timedelta(hours=100))

                # let's have two identical tracks in source without match in sink:
                next(source[0].points()).latitude += 0.06
                source.add(source[0].clone())

                # and two identical tracks in sink without match in source:
                next(sink[1].points()).latitude += 0.07
                sink.add(sink[1].clone())

                # and one track twice in source and once in sink:
                sink.add(source[0])

                # and one track once in source and once in sink:
                sink.add(source[1])

                self.assertBackendLength(source, org_source_len + 1)
                self.assertBackendLength(sink, org_sink_len + 3)

                dump(sink.merge(source, dry_run=True))
                self.assertBackendLength(source, org_source_len + 1)
                self.assertBackendLength(sink, org_sink_len + 3)

                dump(sink.merge(source))
                self.assertBackendLength(source, org_source_len + 1)
                self.assertBackendLength(sink, org_source_len + org_sink_len + 1)

                for _ in range(2):
                    dump(sink.merge(source, remove=True))
                    self.assertBackendLength(source, 0)
                    self.assertBackendLength(sink, org_source_len + org_sink_len)

    @skipIf(*disabled(Directory))
    def test_scan(self):
        """some tests about Backend.scan()."""
        with self.temp_backend(Directory, count=5) as source:
            backend2 = source.clone()
            track = self.create_test_track()
            backend2.add(track)
            self.assertBackendLength(backend2, 6)
            source.scan()  # because it cannot know backend2 added something

    @skipIf(*disabled(Directory))
    def test_lifetrack(self):
        """test life tracking against a local server."""

        def track():
            life = Lifetrack('127.0.0.1', [local_serverdirectory, uplink])
            points = self._random_points(100)
            life.start(points[:50], category=uplink.decode_category(uplink.supported_categories[0]))
            time.sleep(7)
            life.update(points[50:])
            life.end()

        for cls in Backend.all_backend_classes():
            with self.tst_backend(cls):
                with self.temp_backend(Directory) as local_serverdirectory:
                    local_serverdirectory.account.config['id_method'] = 'counter'
                    with self.temp_backend(Directory) as remote_serverdirectory:
                        remote_serverdirectory.account.config['id_method'] = 'counter'
                        with self.lifetrackserver(remote_serverdirectory.url):
                            with self.temp_backend(cls) as uplink:
                                track()
                                track()
                                local_serverdirectory.scan()
                                self.assertBackendLength(local_serverdirectory, 2)
                                local_ids = [x.id_in_backend for x in local_serverdirectory]
                                self.assertEqual(local_ids, ['1', '2'])
                                if cls is TrackMMT:
                                    remote_serverdirectory.scan()
                                    self.assertSameTracks(local_serverdirectory, remote_serverdirectory)
                                elif cls is Mailer:
                                    self.assertEqual(
                                        len(uplink.history), 6,
                                        'Mailer.history: {}'.format(uplink.history))
                                    self.assertIn('Lifetracking starts', uplink.history[0])
                                    self.assertIn('Lifetracking continues', uplink.history[1])
                                    self.assertIn('Lifetracking ends', uplink.history[2])
                                    self.assertIn('Lifetracking starts', uplink.history[3])
                                    self.assertIn('Lifetracking continues', uplink.history[4])
                                    self.assertIn('Lifetracking ends', uplink.history[5])
                                else:
                                    uplink.scan()
                                    self.assertSameTracks(local_serverdirectory, uplink)

    def test_backend_dirty(self):
        """track1._dirty."""
        for cls in Backend.all_backend_classes(needs={'scan', 'write'}):
            with self.tst_backend(cls):
                # category in the Track domain:
                test_category_backend = cls.supported_categories[1]
                test_category = cls.decode_category(test_category_backend)
                with self.temp_backend(cls, count=1) as backend1:
                    track1 = backend1[0]
                    self.assertFalse(track1._dirty)
                    # version 1.1 should perhaps be a test on its own, see Track.xml()
                    track1.category = test_category
                    self.assertEqual(track1.category, test_category)
                    track1._dirty = 'gpx'
                    self.assertEqual(track1.category, test_category)
                    backend2 = backend1.clone()
                    self.assertEqual(backend2[0].category, test_category)
                    b2track = backend2[0]
                    self.assertEqual(b2track.category, test_category)
                    b2track.title = 'another new title'
                    self.assertEqual(b2track.category, test_category)
                    self.assertEqual(backend2[0].category, test_category)
                    track1.title = 'new title'
                    self.assertEqual(track1.category, test_category)
                    backend3 = backend1.clone()
                    self.assertEqual(backend3[0].category, test_category)
                    self.assertFalse(track1._dirty)
                    with track1.batch_changes():
                        track1.title = 'new 2'
                        self.assertEqual(track1._dirty, ['title'])
                    self.assertFalse(track1._dirty)
                    with track1.batch_changes():
                        track1.title = 'new 3'
                        track1.keywords = ['Something', 'something xlse']
                    backend4 = backend1.clone()
                    self.assertEqual(backend4[0].title, 'new 3')

    def test_directory_dirty(self):
        """test gpx._dirty where id_in_backend is not the default.

        Currently track._dirty = 'gpx' changes the file name which is wrong."""

    @skipIf(*disabled(Directory))
    def test_directory(self):
        """directory creation/deletion."""

        with self.temp_backend(Directory) as dir_a:
            self.assertTrue(dir_a.account.is_temporary)
            a_url = dir_a.url
            self.assertTrue(os.path.exists(a_url), a_url)
        self.assertFalse(os.path.exists(a_url), a_url)

        test_url = tempfile.mkdtemp()
        with self.temp_backend(Directory, url=test_url) as dir_b:
            self.assertTrue(dir_b.url == test_url)
        self.assertTrue(os.path.exists(test_url), test_url)
        remove_directory(test_url)

        dir_c = Directory(DirectoryAccount())

        self.assertIn('/gpxity.TestBackends.test_directory_', dir_c.url)
        dir_c.detach()

    @skipIf(*disabled(MMT))
    def test_mmt_empty(self):
        """MMT refuses upload without a specific error message if there is no track point."""
        track = self.create_test_track(MMT)
        del track.gpx.tracks[0]
        with MMT(Account()) as mmt:
            with self.assertRaises(mmt.BackendException):
                mmt.add(track)

    def test_setters(self):
        """For all Track attributes with setters, test if we can change them without changing something else."""
        for cls in Backend.all_backend_classes(needs={'write', 'scan'}):
            with self.tst_backend(cls):
                with self.temp_backend(cls, count=1) as backend:
                    track = backend[0]
                    backend2 = backend.clone()
                    self.assertEqualTracks(track, backend2[0], with_category=False)
                    test_values = {
                        'category': (
                            cls.decode_category(cls.supported_categories[4]),
                            cls.decode_category(cls.supported_categories[2])),
                        'description': ('first description', 'Täst description'),
                        'public': (True, False),
                        'title': ('first title', 'Täst Titel'),
                    }
                    if cls is MMT and backend.subscription == 'free':
                        del test_values['public']
                    if 'keywords' in cls.supported:
                        test_values['keywords'] = (['A', 'Hello Dolly', 'Whatever'], ['Something Else', 'Two'])
                    prev_track = track.clone()
                    for val_idx in (0, 1):
                        for key, values in test_values.items():
                            value = values[val_idx]
                            self.logger.debug('  %s: %s->%s', key, getattr(track, key), value)
                            setattr(track, key, value)
                            setattr(prev_track, key, value)
                            self.assertEqualTracks(prev_track, track)
                            backend2.scan()
                            self.assertEqualTracks(prev_track, backend2[0])

    def test_keywords(self) ->None:
        """Test arbitrary keyword changes.

        Returns:
            None

        """
        def testcases(cls):
            """Prepare test cases for this backend."""
            keywords = {
                backend._encode_keyword(x)
                for x in self._random_keywords(count=50)}
            repeats = 2 if cls.test_is_expensive else 20
            max_size = cls.max_field_sizes.get('keywords', 10000)
            for _ in range(repeats):
                while True:
                    add_keywords = set(random.sample(keywords, random.randint(0, 10)))
                    remove_keywords = set(random.sample(keywords, random.randint(0, 10)))
                    if not add_keywords & remove_keywords:
                        continue
                    expected_keywords = (set(track.keywords) | add_keywords) - remove_keywords
                    if len(', '.join(expected_keywords)) < max_size:
                        break
                yield add_keywords, remove_keywords, expected_keywords

        for cls in Backend.all_backend_classes(needs={'scan', 'keywords', 'write'}):
            with self.tst_backend(cls):
                with self.temp_backend(cls, count=1) as backend:
                    backend2 = backend.clone()
                    track = backend[0]
                    for add_keywords, remove_keywords, expected_keywords in testcases(cls):
                        self.assertEqual(backend._get_current_keywords(track), track.keywords)
                        track.change_keywords(list(add_keywords) * 2)
                        self.assertEqual(
                            backend._get_current_keywords(track),
                            sorted(list(set(track.keywords) | add_keywords)))
                        track.change_keywords('-' + x for x in remove_keywords)
                        self.assertEqual(backend._get_current_keywords(track), sorted(expected_keywords))
                        self.assertEqual(sorted(expected_keywords), sorted(track.keywords))
                        backend2.scan()
                        self.assertEqual(sorted(expected_keywords), backend2[0].keywords)
                    with track.batch_changes():
                        for add_keywords, remove_keywords, expected_keywords in testcases(cls):  # noqa
                            track.change_keywords(add_keywords)
                            track.change_keywords('-' + x for x in remove_keywords)
                    self.assertEqual(sorted(expected_keywords), sorted(track.keywords))
                    backend2.scan()
                    self.assertEqual(
                        backend2._get_current_keywords(backend2[0]),
                        backend2[0].keywords)
                    self.assertEqual(sorted(expected_keywords), backend2[0].keywords)

    @skipIf(*disabled(Directory))
    def test_legal_categories(self):
        """Check if our fixed list of categories still matches the online service."""

        def check():
            """check this backend."""
            downloaded = backend._download_legal_categories()
            self.assertEqual(sorted(backend.supported_categories), downloaded)

        for cls in Backend.all_backend_classes(needs={'own_categories'}):
            if cls is TrackMMT and Directory.is_disabled():
                continue
            with self.tst_backend(cls):
                with self.temp_backend(cls, clear_first=False, cleanup=False) as backend:
                    if cls is TrackMMT:
                        with self.temp_backend(Directory) as serverdirectory:
                            serverdirectory.account.config['id_method'] = 'counter'
                            with self.lifetrackserver(serverdirectory.url):
                                check()
                    else:
                        check()

    def test_long_description(self):
        """Test long descriptions."""
        unlimited_length = 50000  # use this if the backend sets no limit
        for cls in Backend.all_backend_classes(needs={'scan', 'write'}):
            with self.tst_backend(cls):
                with self.temp_backend(cls, count=1) as backend:
                    track = backend[0]
                    max_length = backend._max_length.get('description') or unlimited_length
                    # a backend may encode keywords in description
                    max_descr_length = max_length - (len(backend._encode_description(track)) - len(track.description))
                    track.description = ('long description' * 4000)[:max_descr_length]
                    self.assertEqual(len(backend._encode_description(track)), max_length)
                    clone = backend.clone()[0]
                    self.assertEqual(track.description, clone.description)
                    if max_length < unlimited_length:
                        try_description = 'long description' * 4000
                        encoded = backend._encode_description(track)
                        decoded = backend._decode_description(track.gpx, encoded)
                        self.assertEqual(len(encoded), max_length)
                        self.assertTrue(try_description.startswith(decoded))

    def test_can_encode_all_categories(self):
        """Check if we can encode all internal categories to a given backend value for all backends."""
        for cls in Backend.all_backend_classes(needs={'own_categories'}):
            with self.tst_backend(cls):
                for category in Track.categories:
                    cls.encode_category(category)

    def test_can_decode_all_categories(self):
        """Check if we can decode all backend categories."""
        for cls in Backend.all_backend_classes(needs={'own_categories'}):
            with self.tst_backend(cls):
                for category in cls.supported_categories:
                    cls.decode_category(category)

    def test_category_map(self):
        """Check if all backends can losslessly encode/decode all supported_categories.

        This is done locally assuming that Backend.supported_categories is correct.
        test_legal_categories() tests Backend.supported_categories for correctness.

        """
        for cls in Backend.all_backend_classes(needs={'own_categories'}):
            with self.tst_backend(cls):
                for key, value in cls._category_decoding.items():
                    self.assertIn(key, cls.supported_categories)
                    self.assertIn(value, Track.categories)
                for key, value in cls._category_encoding.items():
                    self.assertIn(key, Track.categories)
                    self.assertIn(value, cls.supported_categories)
                for category in cls.supported_categories:
                    internal = cls.decode_category(category)
                    back = cls.encode_category(internal)
                    self.assertEqual(
                        category, back,
                        '{}: {} -> {} -> {}'.format(cls.__name__, category, internal, back))

    def test_id_change(self):
        """id_in_backend must be legal."""
        for cls in Backend.all_backend_classes(needs={'rename'}):
            if cls is Directory:
                failing = ('', 56, 'a/b', '/', '.', True, False, 5.0)
                working = ('6', 'fdasfds:fasdds')
            elif cls is WPTrackserver:
                failing = ('', 'a', '/', 19, True, False, 5.0, '2147483648')
                working = ('8', '17', '2147483647')
            else:
                raise Exception('untested class {}'.format(cls))
            with self.tst_backend(cls):
                with self.temp_backend(cls) as backend:
                    track = Track()
                    backend.add(track)
                    orig_id = track.id_in_backend
                    for _ in failing:
                        with self.assertRaises(ValueError, msg='for class={} id={}'.format(cls.__name__, _)):
                            track.id_in_backend = _
                    backend.scan()
                    self.assertEqual(len(backend), 1)
                    self.assertEqual(backend[0].id_in_backend, orig_id, 'Testing {} for {}'.format(_, cls.__name__))
                    for _ in working:
                        track.id_in_backend = _
                        backend.scan()
                        self.assertEqual(len(backend), 1)
                        self.assertEqual(backend[0].id_in_backend, _, 'Testing {} for {}'.format(_, cls.__name__))



# -*- coding: utf-8 -*-


# Copyright (c) 2019 Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""Tests for gpxity.backends."""

import unittest
import os
import io
import datetime
import time
import random
from pkgutil import get_data
import tempfile
from contextlib import contextmanager
from subprocess import Popen, PIPE
import logging
try:
    import MySQLdb
    import _mysql_exceptions
except ImportError:
    pass


import gpxpy
from gpxpy.gpx import GPXTrackPoint
from gpxpy.geo import LocationDelta

from ... import Track, Backend, Account, DirectoryAccount
from .. import Mailer, WPTrackserver, Directory, GPSIES, Openrunner, MMT
from ...util import remove_directory
from ...gpx import Gpx

# pylint: disable=attribute-defined-outside-init,protected-access

__all__ = ['BasicTest']


def disabled(*args) ->tuple():
    """True if any of the backends is disabled.

    Returns:
        True or False and reason

    """
    reason = list()
    for _ in args:
        if _.is_disabled():
            reason.append(_.__name__)
    return bool(reason), '{} {} disabled'.format(','.join(reason), 'is' if len(reason) == 1 else 'are')  # noqa


class BasicTest(unittest.TestCase):

    """define some helpers."""

    test_passwd = 'pwd'

    mysql_ip_address = None
    mysql_docker_name = 'gpxitytest_mysql'

    def setUp(self):  # noqa
        """define test specific DirectoryAccount.prefix."""
        self.maxDiff = None  # pylint: disable=invalid-name
        Account.path = os.path.join(os.path.dirname(__file__), 'test_accounts')
        self.logger = logging.getLogger()
        self.logger.level = logging.DEBUG
        self.logger.debug('Using accounts out of %s', Account.path)
        self.start_time = datetime.datetime.now()
        self.unicode_string1 = 'unicode szlig: ß'
        self.unicode_string2 = 'something japanese:の諸問題'
        self.org_dirprefix = DirectoryAccount.prefix
        DirectoryAccount.prefix = 'gpxity.' + '.'.join(self.id().split('.')[-2:]) + '_'  # noqa
        path = tempfile.mkdtemp(prefix=DirectoryAccount.prefix)
        Backend._session.clear()

        DirectoryAccount.prefix = path

        if not Mailer.is_disabled():
            self.start_mailserver()

    def tearDown(self):  # noqa
        """Check if there are still /tmp/gpxitytest.* directories."""
        if not Mailer.is_disabled():
            self.stop_mailserver()
        remove_directory(DirectoryAccount.prefix)
        DirectoryAccount.prefix = self.org_dirprefix
        timedelta = datetime.datetime.now() - self.start_time
        self.logger.debug('%s seconds ', timedelta.seconds)
        logging.shutdown()

    @contextmanager
    def tst_backend(self, backend_cls):  # noqa pylint: disable=arguments-differ
        """With pytest, subTest does not do much. At least print the name."""
        _ = '---------- subTest {}: {} '.format(self.id().split('.')[-1], backend_cls.__name__)
        _ += '-' * (80 - len(_))
        self.logger.debug(_)
        yield super(BasicTest, self).subTest(' ' + backend_cls.__name__)

    @staticmethod
    def _get_track_from_test_file(name: str, backend_cls=None):
        """get data from a predefined gpx file with a random category supported by the backend.

        name is without .gpx

        Returns:=
            A GPX object

        """
        gpx_test_file = os.path.join(os.path.dirname(__file__), '{}.gpx'.format(name))
        if not os.path.exists(gpx_test_file):
            raise Exception('MMTTests needs a GPX file named {}.gpx for testing in {}'.format(
                name, os.getcwd()))
        filename = '{}.gpx'.format(name)
        data = io.StringIO(get_data(__package__, filename).decode('utf-8'))
        result = Track(gpx=Gpx.parse(data))
        if backend_cls:
            result.category = backend_cls.decode_category(random.choice(backend_cls.supported_categories))
        return result

    @classmethod
    def create_test_track(
            cls, backend_class=None, count: int = 1, idx: int = 0, category: str = None, public: bool = False,
            start_time=None, end_time=None):
        """create a :class:`~gpxity.track.Track`.

        It starts off with **test.gpx** and appends a
        last track point, it also changes the time stamp of the last point.
        This is done using **count** and **idx**: The last point is set such that
        looking at the tracks, they all go in a different direction clockwise, with an angle
        in degrees of :literal:`360 * idx / count`.

        Args:
            backend_class: If given, use it as source for a random category
            count: See above. Using 1 as default if not given.
            idx: See above. Using 0 as default if not given.
            category: The wanted value for the track.
                Default: if count == len(:attr:`Track.categories <gpxity.track.Track.categories>`),
                the default value will be backend_class.supported_categories[idx].
                Otherwise a random value from backend_class.supported_categories will be applied.
            public: should the tracks be public or private?
            start_time: If given, assign it to the first point and adjust all following times
            end_time: explicit time for the last point. If None: See above.

        Returns:
            (~gpxity.track.Track): A new track not bound to a backend

        """
        # pylint: disable=too-many-locals
        result = cls._get_track_from_test_file('test')
        if start_time is not None:
            result.adjust_time(start_time - result.first_time)
        last_point = result.last_point()
        if end_time is None:
            end_time = last_point.time + datetime.timedelta(hours=10, seconds=idx)
        new_point = GPXTrackPoint(
            latitude=last_point.latitude, longitude=last_point.longitude + 0.001, time=end_time)
        _ = gpxpy.geo.LocationDelta(distance=1000, angle=360 * idx / count)
        new_point.move(_)
        result.add_points([new_point])

        # now set all times such that they are in order with this track and do not overlap
        # with other test tracks
        _ = result.first_time
        duration = new_point.time - _ + datetime.timedelta(seconds=10)
        for point in result.gpx.walk(only_points=True):
            point.time += duration * idx

        result.title = 'Random GPX # {}'.format(idx)
        result.description = 'Description to {}'.format(result.title)
        if backend_class is None:
            cat_source = Track.categories
        else:
            cat_source = backend_class.supported_categories
            cat_source = [backend_class.decode_category(x) for x in cat_source]
        if category:
            result.category = category
        elif count == len(Track.categories):
            result.category = cat_source[idx]
        else:
            result.category = random.choice(cat_source)
        result.public = public
        return result

    @staticmethod
    def _random_datetime():
        """random datetime between now() - 10 days and now().

        Returns:
            A random datetime

        """
        end = datetime.datetime.now().replace(microsecond=0)
        start = end - datetime.timedelta(days=10)
        delta = end - start
        int_delta = (delta.days * 24 * 60 * 60) + delta.seconds
        random_second = random.randrange(int_delta)
        return start + datetime.timedelta(seconds=random_second)

    @staticmethod
    def _random_keywords(count=100):
        """A set of random keywords, but always the same.

        We do not want to generate too many tag ids for MMT.

        Returns:
            A set of random keywords

        """
        state = random.getstate()
        try:
            random.seed(1)
            basis = 'abcdefghijklmnopqrstuvwxyz'
            basis += basis.upper()
            basis += '/-_+.% $"|\\'
            result = set()
            while len(result) < count:
                candidate = ''.join(random.choice(basis) for x in range(4))
                if candidate[0] not in ' -' and candidate[-1] not in ' ':
                    result.add(candidate)
            return result
        finally:
            random.setstate(state)

    @classmethod
    def _random_points(cls, count=100, root=None):
        """Get some random points.

        Distance between two points will never exceed 200m.

        Returns:
            A list with count points

        """

        start_time = cls._random_datetime()
        if root is None:
            root = GPXTrackPoint(
                latitude=random.uniform(-90.0, 90.0),
                longitude=random.uniform(-180.0, 180.0),
                elevation=0,
                time=start_time)
        result = [root]

        angle = 50
        for _ in range(1, count):
            angle = angle + random.uniform(-20, 20)
            delta = LocationDelta(distance=random.randrange(200), angle=angle)
            point = GPXTrackPoint(latitude=result[-1].latitude, longitude=result[-1].longitude)
            point.move(delta)
            point.elevation = _
            point.time = start_time + datetime.timedelta(seconds=10 * _)
            result.append(point)
        return result

    def assertBackendLength(self, backend, length): # noqa pylint: disable=invalid-name
        """Check length of backend."""
        if len(backend) != length:
            message = ','.join(str(x) for x in backend)
            self.assertEqual(len(backend), length, '{} should have {} tracks: {}'.format(backend, length, message))

    def assertHasKeywords(self, track, expected):  # noqa pylint: disable=invalid-name
        """MMT shows keywords on the website lowercase but internally it capitalizes them."""
        if isinstance(track.backend, MMT):
            original = expected
            expected = []
            for _ in original:
                expected.append(' '.join(x.capitalize() for x in _.split(' ')))
        self.assertEqual(track.keywords, sorted(expected))

    def assertSameTracks(self, backend1, backend2, msg=None, with_category=True, with_last_time=None):  # noqa pylint: disable=invalid-name
        """both backends must hold identical tracks."""
        self.maxDiff = None  # pylint: disable=invalid-name
        if with_last_time is None:
            with_last_time = not (
                isinstance(backend1, (Openrunner, GPSIES)) or isinstance(backend2, (Openrunner, GPSIES)))
        if backend1 != backend2:
            precision = min(backend1.point_precision, backend2.point_precision)
            keys1 = sorted(x.key(with_category, with_last_time, precision=precision) for x in backend1)
            keys2 = sorted(x.key(with_category, with_last_time, precision=precision) for x in backend2)
            self.assertEqual(keys1, keys2, msg)

    def assertEqualTracks(self, track1, track2, msg=None, xml: bool = False, with_category: bool = True):  # noqa pylint: disable=invalid-name
        """both tracks must be identical. We test more than necessary for better test coverage.

        Args:

            xml: if True, also compare xml()"""
        self.maxDiff = None

        # GPSIES: when uploading tracks. GPSIES sometimes assigns new times to all points,
        # starting at 2010-01-01 00:00. Until I find the reason, ignore point times for comparison.
        # Openrunner always does.
        # MMT now seems to convert times between utc and local time. UP- and downloading
        # the same track changes the point times.
        no_time_backend = (GPSIES, Openrunner, MMT)
        with_last_time = not (
            isinstance(track1.backend, no_time_backend) or isinstance(track2.backend, no_time_backend))
        precision = Backend.point_precision
        if track1.backend and track1.backend.point_precision < precision:
            precision = track1.backend.precision
        if track2.backend and track2.backend.point_precision < precision:
            precision = track2.backend.precision
        self.assertEqual(
            track1.key(with_category, with_last_time, precision=precision),
            track2.key(with_category, with_last_time, precision=precision), msg)
        self.assertTrue(track1.points_equal(track2), msg)
        if xml:
            self.assertEqual(track1.gpx.xml(), track2.gpx.xml(), msg)

    def assertNotEqualTracks(self, track1, track2, msg=None, with_category: bool = True):  # noqa pylint: disable=invalid-name
        """both tracks must be different. We test more than necessary for better test coverage."""
        self.assertNotEqual(track1.key(with_category), track2.key(with_category), msg)
        self.assertFalse(track1.points_equal(track2), msg)
        self.assertNotEqual(track1.gpx.xml(), track2.gpx.xml(), msg)

    def assertTrackFileContains(self, track, string, msg=None):  # noqa pylint: disable=invalid-name
        """Assert that string is in the physical file. Works only for Directory backend."""
        with open(track.backend.gpx_path(track.id_in_backend), encoding='utf8') as trackfile:
            data = trackfile.read()
        self.assertIn(string, data, msg)

    def setup_backend(  # pylint: disable=too-many-arguments
            self, cls_, test_name: str = None, url: str = None, count: int = 0,
            clear_first: bool = None, category: str = None,
            public: bool = None):
        """set up an instance of a backend with count tracks.

        If count == len(:attr:`Track.categories <gpxity.track.Track.categories>`),
        the list of tracks will always be identical. For an example
        see :meth:`TestBackends.test_all_category <gpxity.backends.test.test_backends.TestBackends.test_all_category>`.

        Args:
            cls_ (Backend): the class of the backend to be created
            username: use this to for a specific accout name. Default is 'gpxitytest'.
                Special case WPTrackserver: pass the IP address of the mysql test server
            url: for the backend, only for Directory
            count: how many random tracks should be inserted?
            clear_first: if True, first remove all existing tracks. None: do if the backend supports it.
            category: The wanted category, one out of Track.categories. But this is a problem because we do the same
                call for all backend classes and they support different categories. So: If category is int, this is an
                index into Backend.supported_categories which will be decoded into Track.categories

            public: should the tracks be public or private? Default is False.
                Exception: MMT with subscription free has default True

        Returns:
            the prepared Backend

        """
        # pylint: disable=too-many-branches
        if clear_first is None:
            clear_first = 'remove' in cls_.supported
        if isinstance(category, int):
            category = cls_.decode_category(cls_.supported_categories[category])

        if cls_ is Directory:
            account = DirectoryAccount(url)
        else:
            assert url is None
            kwargs = dict()
            if cls_ is WPTrackserver:
                if not self.find_mysql_docker():
                    self.create_temp_mysqld()  # only once for all tests
                self.create_db_for_wptrackserver()  # recreate for each test
                kwargs['Url'] = self.mysql_ip_address
            if test_name:
                account_name = '{}_{}_unittest'.format(cls_.__name__, test_name)
            else:
                account_name = '{}_unittest'.format(cls_.__name__)
            account = Account(account_name, **kwargs)
        result = cls_(account)
        if clear_first and'scan' in cls_.supported and 'write' in cls_.supported:
            result.remove_all()
        if public is None:
            public = cls_ is MMT and result.subscription == 'free'

        if count:
            # if count == 0, skip this. Needed for write-only backends like Mailer.
            while count > len(result):
                track = self.create_test_track(cls_, count, len(result), category=category, public=public)
                result.add(track)
            self.assertGreaterEqual(len(result), count)
            if clear_first:
                self.assertEqual(len(result), count)
        return result

    @staticmethod
    @contextmanager
    def lifetrackserver(directory):
        """Start and ends a server for lifetrack testing."""
        exec_name = 'bin/gpxity_server'
        logfile = os.path.join(directory, 'gpxity_server.log')
        if not os.path.exists(exec_name):
            exec_name = 'gpxity_server'
        cmdline = '{} --loglevel debug --servername localhost --port 12398 {}'.format(
            exec_name, directory)
        user_filename = os.path.join(directory, '.users')
        if not os.path.exists(user_filename):
            with open(user_filename, 'w') as user_file:
                user_file.write('gpxitytest:gpxitytestpw\n')
        process = Popen(cmdline.split(), stdout=open(logfile, 'a'), stderr=open(logfile, 'a'))
        try:
            time.sleep(1)  # give the server time to start
            yield
        finally:
            if os.path.exists(user_filename):
                os.remove(user_filename)
            if os.path.exists(logfile):
                for _ in open(logfile):
                    if 'INFO' not in _ and 'HTTP/1.1' not in _:
                        logging.debug('SRV: %s', _.rstrip())
                os.remove(logfile)
            elif os.path.exists(directory):
                logging.debug('SRV: Directory exists but not gpxity_server.log')
            else:
                logging.debug('SRV: Directory %s does not exist', directory)
            Directory(DirectoryAccount(directory)).remove_all()
            process.kill()

    def start_mailserver(self):
        """Start an smptd server for mail testing."""
        self.mailserver_process = Popen(
            'aiosmtpd -u -n -d -l 127.0.0.1:8025'.split(),
            stdout=open('{}/smtpd_stdout'.format(DirectoryAccount.prefix), 'w'),
            stderr=open('{}/smtpd_stderr'.format(DirectoryAccount.prefix), 'w'))
        time.sleep(1)  # give the server time to start

    def stop_mailserver(self):
        """Stop the smtp server for mail testing."""
        self.mailserver_process.kill()
        for _ in ('out', 'err'):
            filename = '{}/smtpd_std{}'.format(DirectoryAccount.prefix, _)
            if not os.path.exists(filename):
                logging.debug('MAIL: %s not found', filename)
                continue
            if _ == 'err':
                for fileline in open(filename):
                    if not fileline.startswith('INFO:'):
                        logging.debug('MAIL:%s: %s', _, fileline.rstrip())
            os.remove(filename)

    @contextmanager
    def temp_backend(self, cls_, url=None, count=0,  # pylint: disable=too-many-arguments
                     cleanup=True, clear_first=None, category=None,
                     public: bool = None, test_name=None):
        """Just like setup_backend but usable as a context manager. which will call detach() when done."""
        tmp_backend = self.setup_backend(cls_, test_name, url, count, clear_first, category, public)
        try:
            yield tmp_backend
        finally:
            if cleanup and 'remove' in cls_.supported:
                tmp_backend.remove_all()
            tmp_backend.detach()

    @classmethod
    def create_db_for_wptrackserver(cls):
        """Create the mysql database for the WPTrackserver tests."""
        count = 0
        while True:
            try:
                server = MySQLdb.connect(
                    host=cls.mysql_ip_address, user='root', passwd=cls.test_passwd, connect_timeout=2,
                    autocommit=True, charset='utf8',
                    sql_mode='ONLY_FULL_GROUP_BY,STRICT_TRANS_TABLES,NO_ENGINE_SUBSTITUTION')
                break
            except _mysql_exceptions.OperationalError:
                # wait until the docker instance is ready
                count += 1
                if count > 50:
                    raise
                time.sleep(1)
        cursor = server.cursor()
        cursor._defer_warnings = True
        cursor.execute('drop database if exists gpxitytest_db')
        cursor._defer_warnings = False
        cursor.execute('create database gpxitytest_db')
        cursor.execute('use gpxitytest_db')
        cursor.execute("""
            CREATE TABLE wp_ts_locations (
            id int NOT NULL AUTO_INCREMENT,
            trip_id int NOT NULL,
            latitude double NOT NULL,
            longitude double NOT NULL,
            altitude double NOT NULL,
            speed double NOT NULL default 0.0,
            heading double NOT NULL default 0.0,
            updated timestamp NOT NULL DEFAULT CURRENT_TIMESTAMP,
            created timestamp NOT NULL default '0000-00-00 00:00:00',
            occurred timestamp not null default '0000-00-00 00:00:00',
            comment varchar(255) NOT NULL,
            hidden tinyint(1) NOT NULL default 0,
            PRIMARY KEY (id),
            KEY occurred (occurred),
            KEY trip_id (trip_id)
            ) ENGINE=InnoDB
        """)
        cursor.execute("""
            CREATE TABLE wp_ts_tracks (
            id int NOT NULL AUTO_INCREMENT,
            user_id int NOT NULL,
            name varchar({title_length}) NOT NULL,
            updated timestamp NOT NULL DEFAULT CURRENT_TIMESTAMP,
            created timestamp NOT NULL default '0000-00-00 00:00:00',
            source varchar(255) NOT NULL default '',
            comment varchar({descr_length}) NOT NULL default '',
            distance int NOT NULL,
            PRIMARY KEY (id),
            KEY user_id (user_id)
         )
        """.format(
            descr_length=WPTrackserver._max_length['description'],
            title_length=WPTrackserver._max_length['title']))
        cursor.execute("""
            CREATE TABLE wp_users (
            ID bigint(20) unsigned NOT NULL AUTO_INCREMENT,
            user_login varchar(60) NOT NULL DEFAULT '',
            user_pass varchar(255) NOT NULL DEFAULT '',
            user_nicename varchar(50) NOT NULL DEFAULT '',
            user_email varchar(100) NOT NULL DEFAULT '',
            user_url varchar(100) NOT NULL DEFAULT '',
            user_registered datetime NOT NULL DEFAULT '1970-01-01',
            user_activation_key varchar(255) NOT NULL DEFAULT '',
            user_status int NOT NULL DEFAULT '0',
            display_name varchar(250) NOT NULL DEFAULT '',
            PRIMARY KEY (ID),
            KEY user_login_key (user_login),
            KEY user_nicename (user_nicename),
            KEY user_email (user_email)
         )
        """)
        cursor.execute("""
            insert into wp_users (user_login) values(%s)""", ['gpxitytest'])
        cursor.execute("select id,user_login from wp_users")
        server.commit()
        server.close()

    @classmethod
    def find_mysql_docker(cls) ->bool:
        """Find an already running docker.

        Returns:
            the IP address

        """
        if cls.mysql_ip_address is not None:
            return True
        cls.mysql_ip_address = Popen([
            'docker', 'inspect', '--format', '{{range .NetworkSettings.Networks}}{{.IPAddress}}{{end}}',
            cls.mysql_docker_name], stdout=PIPE).communicate()[0].decode().strip()
        if cls.mysql_ip_address == '':
            cls.mysql_ip_address = None
        return bool(cls.mysql_ip_address)

    @classmethod
    def create_temp_mysqld(cls):
        """Create a temporary mysql server and initialize it with test data for WPTrackserver."""
        cmd = [
            'docker', 'run', '--name', cls.mysql_docker_name, '--detach',
            '--env', 'MYSQL_ROOT_PASSWORD={}'.format(cls.test_passwd), 'mysql']
        with Popen(cmd, stdout=PIPE, stderr=PIPE) as _:
            std_err = _.stderr.read().strip().decode('utf8')
            if std_err:
                logging.error(std_err)
                if 'on network bridge: failed to add the host' in std_err:
                    logging.error('did you reboot after kernel upgrade?')
                raise Exception('Cannot run docker: {}'.format(std_err))
        if not cls.find_mysql_docker():
            raise Exception('Unknown problem while creating mysql docker')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) 2019 Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

# PYTHON_ARGCOMPLETE_OK
# for command line argument completion, put this into your .bashrc:
# eval "$(register-python-argcomplete gpxdo)"
# or see https://argcomplete.readthedocs.io/en/latest/


"""lifetrack_client is a command line tool for lifetrack testing, client side."""

import argparse
import os
import sys
import logging
import time
import random

try:
    import argcomplete
    # pylint: disable=unused-import
except ImportError:
    pass

from gpxpy import gpx as mod_gpx
GPXTrack = mod_gpx.GPXTrack
GPXTrackSegment = mod_gpx.GPXTrackSegment


# This uses not the installed copy but the development files
_ = os.path.dirname(sys.path[0] or sys.path[1])
if os.path.exists(os.path.join(_, 'gpxity', '__init__.py')):
    sys.path.insert(0, _)
# pylint: disable=wrong-import-position
from gpxity import Track, Lifetrack, Backend, Directory, MMT, GPSIES, TrackMMT  # noqa


class Main:

    """this is where the work is done."""

    def __init__(self):
        """No args."""
        # pylint: disable=too-many-branches,too-many-nested-blocks

        self.exit_code = 0
        self.options = None
        self.parse_commandline()
        if self.exit_code:
            return
        self.logger = logging.getLogger()
        self.logger.setLevel(self.options.loglevel.upper())
        self.source = None
        try:
            source = Backend.instantiate(self.options.source)
            assert isinstance(source, Track)
            backend = Backend.instantiate(self.options.backend)
            assert isinstance(backend, Backend)
            life = Lifetrack('127.0.0.1', [backend])
            all_points = list(source.points())
            life.start(all_points[:5])
            for point in all_points[5:]:
                time.sleep(random.randrange(10))
                life.update([point])
            time.sleep(random.randrange(10))
            life.end()
        except Exception as _:  # pylint: disable=broad-except
            self.error(_)

    def error(self, msg, exit_code=None):
        """Print the error message.
        Sets the process exit code.
        With --debug, re-raises the exception."""
        self.logger.error(msg)
        self.exit_code = exit_code or 1
        if self.logger.level == logging.DEBUG:
            raise msg

    def parse_commandline(self):
        """into self.options."""
        # pylint: disable=too-many-statements, too-many-branches
        parser = argparse.ArgumentParser('lifetrack_client')
        parser.add_argument('--source', help='the track with test data')
        parser.add_argument('--backend', help='the server')
        parser.add_argument(
            '--loglevel', help='set the loglevel',
            choices=('debug', 'info', 'warning', 'error'), default='error')
        try:
            argcomplete.autocomplete(parser)
        except NameError:
            pass

        if len(sys.argv) < 2:
            parser.print_usage()
            sys.exit(2)

        self.options = parser.parse_args()


sys.exit(Main().exit_code)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) 2019 Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""This module defines :class:`~gpxity.backend.Backend`."""

# pylint: disable=protected-access

import os
from inspect import getmembers, isclass, getmro
import dis
import importlib

from .accounts import Account, DirectoryAccount

__all__ = ['BackendBase']


class BackendBase:

    """Classmethods.

    Track cannot import Backend because Backend imports Track.

    BackendBase is imported by both, containing what both need.

    """

    # pylint: disable=too-many-instance-attributes

    __all_backend_classes = None

    _dirty_separator = '__DIRTY_PAR__'

    class BackendException(Exception):
        """Is raised for general backend exceptions, especially error messages from a remote server"""

    @classmethod
    def _is_implemented(cls, method) ->bool:
        """False if the first instruction in method raises NotImplementedError or if the method does nothing.

        Returns:
            True if method is implemented

        """
        first_instruction = next(dis.get_instructions(method.__code__))
        return first_instruction is not None and first_instruction.argval != 'NotImplementedError'

    @classmethod
    def is_disabled(cls) ->bool:
        """True if this backend is disabled by env variable GPXITY_DISABLE_BACKENDS.

        This variable is a comma separated list of Backend class names.

        Returns:
            True if disabled

        """
        disabled = os.getenv('GPXITY_DISABLE_BACKENDS')
        if not disabled:
            return False
        clsname = cls.__name__.split('.')[-1].lower()
        return clsname in disabled.lower().split()

    @classmethod
    def find_class(cls, name: str):
        """Find the Backend class name "name".

        The only backend where the backend is not specified
        is Directory, so that is returned if no class is found.

        Args:
            name: May be anycase (upper,lower). Must match an
                existing Backend class name.

        Returns:
            the backend class or Exception

        """
        assert name
        for _ in cls.all_backend_classes():
            if _.__name__.lower() == name.lower():
                return _
        raise Exception('find_class failed for {}'.format(name))

    @classmethod
    def _find_local(cls, name: str) ->str:
        """If name refers to a local file, return its expanded path.dirname.

        Returns: The expanded path or None.

        """
        name = os.path.expanduser(name)
        if os.path.exists(name):
            if name.endswith('.gpx'):
                name = name[:-4]
            return name
        if not name.endswith('.gpx'):
            if os.path.exists(name + '.gpx'):
                return name
        dirname, file = os.path.split(name)
        if file:
            if os.path.isdir(dirname):
                return name
        return None

    @classmethod
    def parse_objectname(cls, name):
        """Parse the full identifier for a track.

        1. if name is an existing file or directory, or if name.gpx is an existing file, Backend will be Directory
        2. if ":" is not in name: Backend will be Directory, url=None, track_id=name without .gpx
        3. the part before the first ":" is used as key into accounts. Not case sensitive.

        Args:
            name: the full identifier for a Track

        Returns:
            A tuple with account, track_id

        """
        assert name
        expanded = cls._find_local(name)
        if expanded:
            if os.path.isdir(expanded):
                url = expanded
                track_id = None
            else:
                url, track_id = os.path.split(expanded)
                if not url and not os.path.exists(track_id):
                    url = '.'
            account = DirectoryAccount(url)
        elif ':' not in name:
            url, track_id = os.path.split(name)
            if track_id.endswith('.gpx'):
                track_id = track_id[:-4]
            account = DirectoryAccount(url)
        else:
            _ = name.split(':')
            account_name = _[0]
            track_id = ':'.join(_[1:]) or None
            account = Account(account_name)
            # backend name in accounts is case insensitive, we want the exact name
            account.backend = cls.find_class(account.backend).__name__
        return account, track_id

    @classmethod
    def all_backend_classes(cls, exclude=None, needs=None):
        """Find all backend classes.

        Args:
            exclude: A list with classes to be excluded
            needs: set(str) with needed supported actions

        Returns:
            A sorted list of all backend classes. Disabled backends are not
            returned.

        """
        if cls.__all_backend_classes is None:
            backends_directory = os.path.join(os.path.dirname(__file__), 'backends')
            if not os.path.exists(backends_directory):
                raise Exception('we are not where we should be')
            cls.__all_backend_classes = list()
            mod_names = os.listdir(backends_directory)
            for mod in mod_names:
                if not mod.endswith('.py'):
                    continue
                if mod == '__init__':
                    continue
                try:
                    imported = importlib.import_module('.backends.{}'.format(mod[:-3]), __package__)
                    classes = (x[1] for x in getmembers(imported, isclass))
                    # isinstance and is do not work here
                    classes = [x for x in classes if BackendBase in getmro(x)[1:]]
                    classes = [x for x in classes if not x.is_disabled()]
                    cls.__all_backend_classes.extend(classes)
                except ImportError:
                    pass
            cls.__all_backend_classes = {
                x for x in cls.__all_backend_classes if x.__name__ != 'Backend'}
        if exclude is None:
            exclude = list()
        if needs is None:
            needs = set()
        return sorted(
            (x for x in cls.__all_backend_classes
             if x not in exclude and needs < x.supported), key=lambda x: x.__name__)

    @classmethod
    def _check_id_legal(cls, value):
        """Check if this backend accepts value as id.

        If not, raise ValueError

        """
        if value is not None:
            if not isinstance(value, str):
                raise ValueError('{}: id_in_backend must be str'.format(value))
            if value == '':
                raise ValueError('id_in_backend must not be an empty string')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) 2019 Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""This implements :class:`gpxity.openrunner.Openrunner` for https://www.openrunner.com."""

# pylint: disable=protected-access

from html.parser import HTMLParser
import datetime
from collections import defaultdict

import logging

import requests
from gpxpy import gpx as mod_gpx


from .. import Backend
from ..gpx import Gpx

GPXTrackPoint = mod_gpx.GPXTrackPoint

if False:  # pylint: disable=using-constant-test
    try:
        import http.client as http_client
    except ImportError:
        # Python 2
        import httplib as http_client
    http_client.HTTPConnection.debuglevel = 1

__all__ = ['Openrunner']


class OpenrunnerRawTrack:

    """raw data from the gpies html page."""

    # pylint: disable=too-few-public-methods
    def __init__(self):
        """See class docstring."""
        self.track_id = None
        self.title = None
        self.time = None
        self.distance = None
        self.category = None


class ParseOpenrunnerCategories(HTMLParser):  # pylint: disable=abstract-method

    """Parse the legal values for category from html."""

    def __init__(self):
        """See class docstring."""
        super(ParseOpenrunnerCategories, self).__init__()
        self.result_names = ['Other']
        self.result_ids = [2]
        self.current_tag = None
        self.seeing_list = False

    def feed(self, data):
        """get data."""
        self.seeing_list = False
        super(ParseOpenrunnerCategories, self).feed(data)

    def handle_starttag(self, tag, attrs):
        """starttag from the parser."""
        self.current_tag = tag
        attributes = defaultdict(str)
        for key, value in attrs:
            attributes[key] = value
        if tag == 'select' and attributes['id'] == 'search_activity':
            self.seeing_list = True
        if self.seeing_list and tag == 'option':
            if attributes['value'] != '':
                self.result_ids.append(int(attributes['value']))

    def handle_endtag(self, tag):
        """handle end of track list."""
        if tag == 'select':
            self.seeing_list = False

    def handle_data(self, data):
        """data from the parser."""
        if self.seeing_list:
            data = data.strip()
            if not data:
                return
            if self.current_tag == 'option':
                if data != 'Select an activity':
                    self.result_names.append(data)


class ParseOpenrunnerSubscription(HTMLParser):  # pylint: disable=abstract-method

    """Parse the current subscription from html."""

    def __init__(self):
        """See class docstring."""
        super(ParseOpenrunnerSubscription, self).__init__()
        self.category = None
        self.seeing_card_title = False
        self.current_card_title = None
        self.result = None

    def handle_starttag(self, tag, attrs):
        """starttag from the parser."""
        attributes = dict(attrs)
        self.seeing_card_title = False
        if attributes.get('class') == 'card-title':
            self.seeing_card_title = True
            self.current_card_title = None

    def handle_data(self, data):
        """Handle data."""
        data = data.strip()
        if data == '':
            return
        if self.seeing_card_title:
            self.current_card_title = data
        if data == 'Subscription in progress':
            self.result = self.current_card_title


class ParseOpenrunnerActivity(HTMLParser):  # pylint: disable=abstract-method

    """Parse the category value for a track from html."""

    def __init__(self):
        """See class docstring."""
        super(ParseOpenrunnerActivity, self).__init__()
        self.current_tag = None
        self.category = None

    def handle_starttag(self, tag, attrs):
        """starttag from the parser."""
        self.current_tag = tag
        attributes = defaultdict(str)  # TODO: class MyHTMLParser with attributes getter
        for key, value in attrs:
            attributes[key] = value
        attributes = dict(attrs)
        if tag == 'input' and attributes['name'] == 'trackTypes' and 'checked' in attributes:
            self.category = attributes['id']


class ParseOpenrunnerList(HTMLParser):  # pylint: disable=abstract-method

    """get some attributes available only on the web page.

    Of course, this is highly unreliable. Just use what we can get."""

    def __init__(self):
        """See class docstring."""
        super(ParseOpenrunnerList, self).__init__()
        self.result = dict()
        self.result['tracks'] = list()
        self.track = None
        self.column = 0
        self.current_tag = None
        self.seeing_list = False
        self.after_list = False
        self.seeing_a = False

    def feed(self, data):
        """get data."""
        self.track = None
        self.column = 0
        self.current_tag = None
        self.seeing_list = False
        self.after_list = False
        self.seeing_a = False
        super(ParseOpenrunnerList, self).feed(data)

    def handle_starttag(self, tag, attrs):
        """starttag from the parser."""
        self.current_tag = tag
        attributes = defaultdict(str)
        for key, value in attrs:
            attributes[key] = value
        if tag == 'tbody':
            self.seeing_list = True
        if not self.seeing_list:
            return
        if tag == 'tr':
            self.track = OpenrunnerRawTrack()
            self.column = 0
            self.seeing_a = False
        elif tag == 'td':
            self.column += 1
        elif self.after_list and tag == 'a':
            self.seeing_a = True
            value = attributes['value'].strip()

    def handle_endtag(self, tag):
        """handle end of track list."""
        if tag == 'tbody':
            self.seeing_list = False
            self.after_list = True

    def handle_data(self, data):
        """data from the parser."""
        data = data.strip()
        if not data:
            return
        if self.seeing_list:
            if self.column == 2:
                self.track.track_id = data
            elif self.column == 3:
                if self.current_tag == 'h6' and self.track.title is None:
                    self.track.title = data
            elif self.column == 4:
                self.track.category = data
            elif self.column == 7:
                self.track.distance = float(data)
            elif self.column == 10:
                self.track.time = datetime.datetime.strptime(data, '%d-%m-%Y')
                self.result['tracks'].append(self.track)


class Openrunner(Backend):

    """The implementation for openrunner.com.

    The track ident is the ID given by openrunner.

    Searching arbitrary tracks is not supported. Openrunner only looks at the
    tracks of a specific user.

    Args:
        account (:class:`~gpxity.accounts.Account`): The account to be used.
            Alternatively a dict can be passed to build an ad hoc :class:`~gpxity.accounts.Account`
            instance.

    """

    # pylint: disable=abstract-method

    max_field_sizes = {'keywords': 200}

    _default_description = 'None yet. Let everyone know how you got on.'

    point_precision = 5

    supported_categories = (
        'Cycling - Road',
        'Canoe-Kayak',
        'Cycling - Gravel',
        'Cycling - MTB',
        'Cycling - Touring',
        'Footbiking',
        'Hiking',
        'Horse riding',
        'Longboard',
        'Nordic walking',
        'River navigation',
        'Rollerblading',
        'Running - Trail',
        'Running - Urban Trail',
        'Running- Road',
        'Skiing - Backcountry',
        'Skiing - Crosscountry',
        'Skiing - Rollerskiing',
        'Skiing - Touring',
        'Snowshoeing',
        'Stand Up Paddle',
        'Swimming',
        'Swimrun',
        'Walking',
        'Other'
    )

    _category_decoding = {
        'Canoe-Kayak': 'Canoeing',
        'Footbiking': 'Cycling - Foot',
        'Longboard': 'Miscellaneous',
        'Other': 'Miscellaneous',
        'River navigation': 'Miscellaneous',
        'Rollerblading': 'Skating - Inline',
        'Running - Trail': 'Running',
        'Running - Urban Trail': 'Running',
        'Running- Road': 'Running - Road',
        'Skiing - Rollerskiing': 'Skiing - Roller',
        'Stand Up Paddle': 'Stand up paddle boarding',
        'Swimrun': 'Swimming',
    }

    # translate internal names to Openrunner names
    _category_encoding = {
        'Cabriolet': 'Other',
        'Canoeing': 'Canoe-Kayak',
        'Coach': 'Other',
        'Crossskating': 'Other',
        'Cycling': 'Cycling - Road',
        'Cycling - Foot': 'Footbiking',
        'Cycling - Hand': 'Cycling - Road',
        'Cycling - Indoor': 'Other',
        'Driving': 'Other',
        'Enduro': 'Other',
        'Flying': 'Other',
        'Geocaching': 'Other',
        'Gliding': 'Other',
        'Hang gliding': 'Other',
        'Hiking - Speed': 'Hiking',
        'Hot air ballooning': 'Other',
        'Jet skiing': 'Other',
        'Kayaking': 'Canoe-Kayak',
        'Kiteboarding': 'Other',
        'Miscellaneous': 'Other',
        'Motor racing': 'Other',
        'Motorcycling': 'Other',
        'Motorhome': 'Other',
        'Mountaineering': 'Other',
        'Nordic walking': 'Other',
        'Off road driving': 'Other',
        'Orienteering': 'Other',
        'Pack animal trekking': 'Horse riding',
        'Paragliding': 'Other',
        'Pedelec': 'Other',
        'Powerboating': 'River navigation',
        'Rowing': 'River navigation',
        'Running': 'Running- Road',
        'Running - Road': 'Running- Road',
        'Sailing': 'Other',
        'Sea kayaking': 'Canoe-Kayak',
        'Sightseeing': 'Other',
        'Skateboarding': 'Other',
        'Skating': 'Other',
        'Skating - Inline': 'Rollerblading',
        'Skiing': 'Skiing - Touring',
        'Skiing - Alpine': 'Other',
        'Skiing - Nordic': 'Skiing - Backcountry',
        'Skiing - Touring': 'Skiing - Backcountry',
        'Snowboarding': 'Other',
        'Snowshoeing': 'Other',
        'Stand up paddle boarding': 'Stand Up Paddle',
        'Train': 'Other',
        'Wheelchair': 'Other',
        'Windsurfing': 'Other',
        'Wintersports': 'Other',
    }

    # translate Openrunner names to Openrunner numbers
    _legal_categories_numbers = {
        'Canoe-Kayak': 18,
        'Cycling - Gravel': 20,
        'Cycling - MTB': 3,
        'Cycling - Road': 1,
        'Cycling - Touring': 11,
        'Footbiking': 13,
        'Hiking': 9,
        'Horse riding': 4,
        'Longboard': 25,
        'Nordic walking': 12,
        'Other': 2,
        'River navigation': 8,
        'Rollerblading': 5,
        'Running - Trail': 10,
        'Running - Urban Trail': 19,
        'Running- Road': 21,
        'Skiing - Backcountry': 14,
        'Skiing - Crosscountry': 15,
        'Skiing - Rollerskiing': 16,
        'Skiing - Touring': 22,
        'Snowshoeing': 17,
        'Stand Up Paddle': 24,
        'Swimming': 6,
        'Swimrun': 23,
        'Walking': 7,
    }

    assert set(_legal_categories_numbers.keys()) == set(supported_categories)

    default_url = 'https://www.openrunner.com'

    @staticmethod
    def __encode_number(nbr) -> str:
        """Encode a single unsigned number.

        Returns: the encoded string

        """
        result = ''
        while nbr >= 32:
            result += chr(95 + (nbr & 31))
            nbr >>= 5
        result += chr(63 + nbr)
        return result

    @classmethod
    def __encode_signed_number(cls, nbr) ->str:
        """Encode a single signed number.

        Returns: The encoded string

        """
        tmp = nbr << 1
        if nbr < 0:
            tmp = ~tmp
        return cls.__encode_number(tmp)

    @classmethod
    def _encode_points(cls, points) ->str:
        """Encode a list of points.

        Returns: the encoded string

        """
        result = ''
        prev_lat = 0
        prev_lon = 0
        for point in points:
            lat = round(point.latitude * 100000)
            lon = round(point.longitude * 100000)
            delta_lat = lat - prev_lat
            delta_lon = lon - prev_lon
            prev_lat = lat
            prev_lon = lon
            result += cls.__encode_signed_number(delta_lat)
            result += cls.__encode_signed_number(delta_lon)
        return result

    @staticmethod
    def _decode_points(input_str) ->list:
        """Decode str into a list of points.

        Returns: list(GPXTrackPoint)

        """
        def decode_number():
            """Decode a single number."""
            nonlocal input_str
            result = 0
            shift = 0
            while True:
                ord_value = ord(input_str[0]) - 63
                result |= (31 & ord_value) << shift
                input_str = input_str[1:]
                if ord_value < 32:
                    break
                shift += 5
            if 1 & result:
                return ~(result >> 1)
            return result >> 1

        def blow_up(nbr):
            return round(0.00001 * nbr, 5)
        result = list()
        latitude = longitude = 0
        while input_str:
            latitude += decode_number()
            longitude += decode_number()
            result.append(GPXTrackPoint(latitude=blow_up(latitude), longitude=blow_up(longitude)))
        return result

    def _download_legal_categories(self):
        """Needed only for unittest.

        Returns: list(str)
            all legal values for category.

        """
        response = self.__get(action='route/search/page')
        category_parser = ParseOpenrunnerCategories()
        category_parser.feed(response.text)
        return sorted(category_parser.result_names)

    @property
    def subscription(self) ->str:
        """Get the subscription model.

        Returns: The name of the subscription.

        """
        if self._cached_subscription is None:
            parser = ParseOpenrunnerSubscription()
            parser.feed(self.__get(action='user/mysubscription').text)
            if parser.result == 'Standard':
                self._cached_subscription = 'free'
            else:
                self._cached_subscription = 'full'
            self.logger.debug('%s: subscription: %s', self.account, self._cached_subscription)
        return self._cached_subscription

    @property
    def session(self):
        """The requests.Session for this backend. Only initialized once.

        Returns:
            The session

        """
        ident = str(self)
        if ident not in self._session:
            if not self.account.username:
                raise self.BackendException('{}: Needs authentication data'.format(self.url))
            self._session[ident] = requests.Session()
            if self.account.password:
                data = {
                    'language': 'en',
                    'login': self.account.username,
                    'password': self.account.password,
                }
                self._session[ident].response = self._session[ident].post(
                    '{}/user/login'.format(self.url),
                    data=data, timeout=self.timeout)
                self._check_response(self._session_response, data)
        if self._session_response is None:
            self.logger.info('Openrunner.session got no _session_response')
        return self._session[ident]

    @property
    def _session_response(self):
        """The last response received.

        Returns: The response

        """
        # TODO: also in GPSIES
        ident = str(self)
        if ident in self._session:
            if hasattr(self._session[ident], 'response'):
                return self._session[ident].response
        return None

    def __http_post(self, post_type, action: str, data=None):
        """Common code for HTTP POST.

        Returns: the response

        """
        if data is None:
            data = dict()
        data['_'] = int(datetime.datetime.now().timestamp())
        full_url = '{}/{}'.format(self.url, action)
        self.session  # because headers needs accessToken  pylint: disable=pointless-statement
        headers = {'X-Language': 'en'}
        if self._session_response:
            headers['Authorization'] = 'Bearer {}'.format(self._session_response.json()['user']['accessToken'])
            method = getattr(self.session, post_type)
        else:
            method = getattr(requests, post_type)
        response = method(full_url, data=data, headers=headers, timeout=self.timeout)
        self._check_response(response, data)
        return response

    def __get(self, action: str, data=None):
        """common code for a GET within the session.

        Returns: the response

        """
        return self.__http_post("get", action, data)

    def __post(self, action: str, data):
        """common code for a POST within the session.

        Returns: the response

        """
        return self.__http_post("post", action, data)

    def __delete(self, action: str, data):
        """common code for a POST within the session.

        Returns:
            the response

        """
        return self.__http_post("delete", action, data)

    @classmethod
    def decode_category(cls, value: str) ->str:
        """Translate the value from Openrunner into internal one.

        Returns:
            The decoded name

        """
        try:
            return super(Openrunner, cls).decode_category(value)
        except Exception:
            if value == 'Autre':  # TODO: Should not happen
                logging.error("Openrunner said Autre")
                return 'Miscellaneous'
            reverse_nbr = dict(zip(cls._legal_categories_numbers.values(), cls._legal_categories_numbers.keys()))
            try:
                if int(value) in reverse_nbr:
                    return cls.decode_category(reverse_nbr[int(value)])
            except ValueError:
                pass
            raise cls.BackendException('Openrunner gave us an unknown track type {}'.format(value))

    def _load_track_headers(self):
        """get all tracks for this user."""
        if self.account.password:
            response = self.__get(action='user/myroute')
        else:
            response = self.__get(action='route/findby?author={}'.format(self._get_author()))
        page_parser = ParseOpenrunnerList()
        page_parser.feed(response.text)
        for raw_data in page_parser.result['tracks']:
            track = self._found_track(raw_data.track_id)
            track.title = raw_data.title
            track.time = raw_data.time
            if raw_data.distance:
                track.distance = raw_data.distance
            if raw_data.category:
                track.category = self.decode_category(raw_data.category)

    def _read_all(self, track):
        """Get the entire track."""
        response = self.__get(action='route/{}'.format(track.id_in_backend))
        route = response.json()['route']
        points = self._decode_points(route['shape']['full_encoded'])
        gpx = Gpx()
        gpx.add_points(points)
        gpx.name = route['name']
        gpx.description = route['description']
        gpx.keywords = route['keyword']
        track.gpx = gpx
        # the date format seems to depend on the language. fr would be %d-%m-%Y
        track.time = datetime.datetime.strptime(route['updatedDate'], '%Y/%m/%d')
        track.public = not route['private']
        track.category = self.decode_category(route['activity'])

    def _check_response(self, response, data):
        """are there error messages?."""
        if response.status_code != 200:
            if 'route_activity_id_foreign' in response.text:
                raise self.BackendException('Category id is illegal: {}'.format(data))
            if 'route.keyword may not be greater' in response.text:
                raise self.BackendException('Keywords too long: {}'.format(data))
            if 'Unauthorized' in response.text:
                raise self.BackendException('{}: Needs authentication data'.format(self.url))
            raise self.BackendException(response.text)

    def _write_all(self, track) ->str:
        """save full gpx track at Openrunner.

        Returns:
            The new id_in_backend

        """
        if not track.gpx.get_track_points_no():
            raise self.BackendException('Openrunner does not accept a track without trackpoints:{}'.format(track))
        points = list(track.points())
        data = {
            'route[activity]': self._legal_categories_numbers[self.encode_category(track.category)],
            'route[description]': track.description,
            'route[elevation][sampleEncoded]': self._encode_points(points),
            'route[elevation][sampleIntervalInMeter]': track.distance / len(points),
            'route[end][lat]': points[-1].latitude,
            'route[end][lng]': points[-1].longitude,
            'route[is_private]': 0 if track.public else 1,
            'route[is_tested]': 1,
            'route[keyword]': ', '.join(track.keywords),
            'route[labelColor]': '#ffffff',
            'route[lengthInMeter]': min(track.distance, 1200) * 1000,  # TODO: unittest: is max length still 1200km?
            'route[name]': track.title,
            'route[official]': 0,
            'route[shape][pointShapeEncoded]': '',
            'route[shape][pointShapeReducedEncoded]': self._encode_points(points[:20]),
            'route[shape][pointWaypointEncoded]': self._encode_points(points),
            'route[shape][pointWaypointType]': 'A' * len(points),
            'route[shape][showMilestone]': 1,
            'route[shape][strokeColor]': "#b71c0c",
            'route[shape][strokeOpacity]': 0.8,
            'route[shape][strokeWidth]': 5,
            'route[source]': 'openrunner-web',
            'route[start][lat]': points[0].latitude,
            'route[start][lng]': points[0].longitude,
            'route[surface]': 0,
            'route[terrain]': 0,
            'route[waymark]': 0,
        }
        response = self.__post(action='route', data=data)
        json = response.json()
        new_ident = str(json['id'])
        if not new_ident:
            raise self.BackendException('No id found in response')
        old_ident = track.id_in_backend
        track.id_in_backend = new_ident
        if old_ident:
            self._remove_ident(old_ident)
        track.id_in_backend = new_ident
        return new_ident

    def _remove_ident(self, ident: str):
        """remove on the server."""
        self.__delete(action='route', data={'routeIds[0]': ident})

    def detach(self):
        """also close session."""
        super(Openrunner, self).detach()
        if self.session:
            self.session.close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) 2019 Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""This module defines :class:`~gpxity.track.Track`."""

# pylint: disable=protected-access

import datetime
from functools import total_ordering
from contextlib import contextmanager
import weakref
from copy import deepcopy
import logging

# pylint: disable=too-many-lines

from gpxpy import gpx as mod_gpx
from gpxpy.geo import simplify_polyline

from .gpx import Gpx
from .backend_base import BackendBase
from .util import repr_timespan

GPXTrack = mod_gpx.GPXTrack
GPXTrackSegment = mod_gpx.GPXTrackSegment
GPXXMLSyntaxException = mod_gpx.GPXXMLSyntaxException


__all__ = ['Track']


@total_ordering
class Track:  # pylint: disable=too-many-public-methods

    """Represents a track.

    A Track is essentially a GPX file.  A GPX file may contain multiple tracks but whenever
    this documentation says track or Track it does not refer to one of possibly multiple entities in
    the GPX file. It refers to this class Track.

    If a :class:`~gpxity.backend.Backend` supports attributes not directly
    supported by the GPX format like the MapMyTracks track type, they will
    transparently be encodeded in existing GPX fields like keywords, see :attr:`keywords`.

    If a :class:`~gpxity.backend.Backend` (like :class:`~gpxity.backends.wptrackserver.WPTrackserver`)
    does not support keywords, the will transparently be encoded in the description.

    The GPX part is done by https://github.com/tkrajina/gpxpy.

    If a track is assigned to a backend, all changes will by default be written directly to the backend.
    Some backends are able to change only one attribute with little time overhead, others always have
    to rewrite the entire track. Assigning a track with :literal:`gpx == None` will raise an Exception.

    You can use the context manager :meth:`batch_changes`. This holds back updating the backend until
    leaving the context.

    If you manipulate the gpx directly, this goes unnoticed to the updating mechanism. Use :meth:`rewrite` when done.

    Not all backends support everything, you could get the exception NotImplementedError.

    The data will only be loaded from the backend when it is needed. Backends have two ways
    of loading data: Either load a list of Tracks or load all information about a specific track. Often
    loading the list of tracks gives us some attributes for free, so listing
    those tracks may be much faster if you do not want everything listed.

    Absolutely all attributes (like :attr:`title`, :attr:`distance`) are encoded in :attr:`gpx`.
    However you can always assign values to them even if :attr:`gpx` is None. As soon
    as :attr:`gpx` is given, it will be updated.

    All points are always rounded to 6 decimal digits when they are added to the track.
    However some backends may support less than 6 decimals. You can query Backend.point_precision.

    Args:
        gpx (Gpx): Initial content. Can be used if you create a new Track from scratch without
            loading it from some backend.

    Attributes:
        categories (tuple(str)): The legal values for :attr:`~Track.category`. The first one is used
            as default value. This is a superset of the values for the different backends.
            Every backend maps from its internal values into those when reading and maps them
            back when writing. Since not all backends support all values defined here and since
            some backends may define more values than we know, information may get lost when
            converting.

    """

    # pylint: disable = too-many-instance-attributes,too-many-public-methods

    class CannotMerge(Exception):
        """Is raised if :meth:`Track.merge() <gpxity.track.Track.merge>` fails."""

    categories = (
        'Cycling', 'Cycling - Road', 'Cycling - Gravel', 'Cycling - MTB', 'Cycling - Indoor', 'Cycling - Hand',
        'Cycling - Touring', 'Cycling - Foot',
        'Running', 'Running - Trail', 'Running - Urban Trail', 'Running - Road',
        'Sailing', 'Walking', 'Hiking', 'Hiking - Speed',
        'Swimming', 'Driving', 'Off road driving', 'Motor racing', 'Motorcycling', 'Enduro',
        'Skiing', 'Skiing - Touring', 'Skiing - Backcountry', 'Skiing - Crosscountry', 'Skiing - Nordic',
        'Skiing - Alpine', 'Skiing - Roller',
        'Canoeing', 'Kayaking', 'Sea kayaking', 'Stand up paddle boarding',
        'Rowing', 'Windsurfing', 'Kiteboarding', 'Orienteering', 'Mountaineering', 'Skating',
        'Skateboarding', 'Horse riding', 'Hang gliding', 'Gliding', 'Flying', 'Snowboarding',
        'Paragliding', 'Hot air ballooning', 'Nordic walking', 'Snowshoeing', 'Jet skiing', 'Powerboating',
        'Swimrun',
        'Pedelec', 'Crossskating', 'Motorhome', 'Cabriolet', 'Coach',
        'Pack animal trekking', 'Train', 'Wheelchair', 'Sightseeing', 'Geocaching', 'Longboard',
        'River navigation',
        'Skating - Inline',
        'Wintersports',
        'Miscellaneous')

    _obsolete_categories = {
        'Mountain biking': 'Cycling - MTB',
    }

    def __init__(self, gpx=None):
        """See class docstring."""
        if gpx is None:
            gpx = Gpx()
        assert isinstance(gpx, Gpx)
        self.__dirty = list()
        self._batch_changes = False
        self.__ids = list()  # TODO: remove
        self.__id_in_backend = None
        self.__backend = None
        self.__gpx = None
        self.__backend = None
        self.__cached_time = None
        self.__cached_distance = None

        # __header_cache holds all attributes that are set while gpx is None.
        # After gpx is given, if an attribute is not set in gpx, remove it from __header_cache
        # and write it into gpx. Do that for all values from __header_cache before writing
        # to the backend.
        self.__header_cache = dict()

        self._similarity_others = weakref.WeakValueDictionary()
        self._similarities = dict()
        self.__without_fences = None  # used by context manager "fenced()"
        self.gpx = gpx

    def __decode_gpx(self):
        """Extract attributes from gpx.

        This is done whenever gpx changes.

        We have 3 groups of attributes depending on how often
        they are needed an how expensive their computation is,
        in ascending order:

        1. always read from and write to gpx. Examples: title, description
        2. cache now. Examples: category ??? Wirklich???? Oder nur 1. und 3.?
        3. lazy: compute and cache only when needed. Examples: distance, time
            Those are set to None here.

        If an attribute is changed, the full track is always loaded first.

        """
        self.__gpx.decode()

        # lazy attributes:
        self.__cached_distance = None
        self.__cached_time = None
        self._clear_similarity_cache()

    def __encode_gpx(self):
        """Put values into gpx. See __decode_gpx."""
        self.__gpx.encode()
        if self.backend is not None:
            self.__gpx.category = self.backend.encode_category(self.__gpx.category)

    @property
    def backend(self):
        """The backend this track lives in. If the track was constructed in memory, backend is None.

        This is a read-only property. It is set with :meth:`Backend.add <gpxity.backend.Backend.add>`.

        It is not possible to decouple a track from its backend, use :meth:`clone()`.

        Returns:
            The backend

        """
        return self.__backend
        # :attr:`Track.id_in_backend <gpxity.track.Track.id_in_backend>`.

    @property
    def id_in_backend(self) ->str:
        """Every backend has its own scheme for unique track ids.

        Some backends may change this if the track data changes.

        Some backends support assigning a new value. Those are
        currently :class:`~gpxity.backends.directory.Directory` and
        :class:`~gpxity.backends.wptrackserver.WPTrackserver`.
        The others will raise NotImplementedError.

        See also :meth:`Backend.add() <gpxity.backend.Backend.add>`.


        Returns:
            the id in the backend

        """
        return self.__id_in_backend

    @id_in_backend.setter
    def id_in_backend(self, value: str) ->None:
        """Change the id in the backend.

        Illegal changes raise ValueError.

        Args:
            value: The new value

        """
        BackendBase._check_id_legal(value)
        if self.__backend:
            self.__backend._check_id_legal(value)
        if self.__id_in_backend == value:
            return
        if self.__id_in_backend:
            self.__ids.insert(0, str(self))
        if self.__is_decoupled:
            # internal use
            self.__id_in_backend = value
        else:
            if not self.__id_in_backend:
                raise ValueError('Cannot set id_in_backend for yet unsaved track {}'.format(self))
            if not value:
                raise ValueError('Cannot remove id_in_backend for saved track {}'.format(self))
            with self._decouple():
                self.backend._change_ident(self, value)

    def _set_backend(self, value):
        """To be used only by backend implementations."""
        assert self.__is_decoupled
        if self.__backend:
            self.__backend._check_id_legal(self.id_in_backend)
        old_backend = self.__backend
        self.__backend = value
        if self.__gpx.keywords:
            if old_backend is None or old_backend.__class__ != value.__class__:
                # encode keywords for the new backend
                # TODO: unittest
                self.__gpx.encode()
                self.change_keywords(self.__gpx.real_keywords)

    def rewrite(self) ->None:
        """Call this after you directly manipulated  :attr:`gpx`."""
        if not self.__gpx.is_complete:
            raise Exception('Track.rewrite: The track must already be loaded fully')
        self._dirty = 'gpx'

    @property
    def _dirty(self) ->list:
        """
        Check if  the track is in sync with the backend.

        Setting :attr:`_dirty` will directly write the changed data into the backend.

        :attr:`_dirty` can receive an arbitrary string like 'title'. If the backend
        has a method _write_title, that one will be called. Otherwise the
        entire track will be written by the backend.

        Returns:
            list: The names of the attributes currently marked as dirty.

        """
        return self.__dirty

    @_dirty.setter
    def _dirty(self, value):
        """See dirty.getter."""
        if not isinstance(value, str):
            raise Exception('_dirty only receives str')
        if value == 'gpx':
            if self.__without_fences is not None:
                raise Exception(
                    '{}: You may not modify gpx while being in the context manager "fenced()"'.format(self))
            self.__decode_gpx()
        if not self.__is_decoupled:
            self.__dirty.append(value)
            if not self._batch_changes:
                self._rewrite()

    def _clear_similarity_cache(self):
        """Clear similarities cache, also in the other similar tracks."""
        for other in self._similarity_others.values():
            del other._similarity_others[id(self)]
            # TODO: unittest where other does not exist anymore
            del other._similarities[id(self)]
        self._similarity_others = weakref.WeakValueDictionary()
        self._similarities = dict()

    def _clear_dirty(self):
        """To be used by the backend when saving."""
        self.__dirty = list()

    def clone(self):
        """Create a new track with the same content but without backend.

        Returns:
            ~gpxity.track.Track: the new track

        """
        self._load_full()
        result = Track(gpx=self.gpx.clone())
        if self.backend is not None:
            result.__ids.insert(0, str(self))
        return result

    def _rewrite(self):
        """Rewrite all changes in the associated backend.

        If any of those conditions is met, do nothing:

        - we are currently loading from backend: Avoid recursion
        - batch_changes is active
        - we have no backend

        Otherwise the backend will save this track.

        """
        if 'gpx' in self.__dirty:
            self.__decode_gpx()
        if self.backend is None:
            self._clear_dirty()
        if not self.__dirty:
            return
        if 'write' not in self.backend.supported:
            # TODO: unittest
            raise Exception('Rewriting {}: "write" is not supported'.format(self))
        if not self.__is_decoupled and not self._batch_changes:
            with self._decouple():
                old_gpx_keywords = self.__gpx.keywords
                try:
                    self.__encode_gpx()
                    self.backend._rewrite(self, self.__dirty)
                finally:
                    self.__gpx.keywords = old_gpx_keywords
            self._clear_dirty()

    def remove(self):
        """Remove this track in the associated backend.

        If the track is not coupled with a backend, raise an Exception.

        """
        if self.backend is None:
            raise Exception('{}: Removing needs a backend'.format(self))
        self.backend.remove(self.id_in_backend)

    @property
    def first_time(self) ->datetime.datetime:
        """datetime.datetime: start time of track.

        For a simpler implementation of backends, notably :class:`~gpxity.backends.mmt.MMT`
        we ignore gpx.time. Instead we return the time of the earliest track point.
        Only if there is no track point, return gpx.time. If that is unknown
        too, return None.

        For the same reason time is readonly.

        We assume that the first point comes first in time and the last
        point comes last in time. In other words, points should be ordered
        by their time.

        """
        if self.__cached_time is None:
            self.__cached_time = self.__gpx.first_time
            if self.__cached_time is None:
                self.__cached_time = self.gpx.first_time
        return self.__cached_time

    @property
    def distance(self) ->float:
        """For me, the earth is flat.

        This property can only be set while the full track has not yet
        been loaded. The setter is used by the backends when scanning for all tracks.

        Returns:
            the distance in km, rounded to m. 0.0 if not computable.

        """
        if self.__cached_distance is None:
            self.__cached_distance = self.gpx.distance
        return self.__cached_distance

    @distance.setter
    def distance(self, value):
        """The setter."""
        if self.__gpx.is_complete:
            raise Exception('Setting Track.distance is only allowed while the full track has not yet been loaded')
        self.__cached_distance = value

    @property
    def title(self) -> str:
        """str: The title.

        Returns:
            the title

        """
        if self.__gpx.name == Gpx.undefined_str:
            self._load_full()
            if self.__gpx.name == Gpx.undefined_str:
                self.__gpx.name = ''
        return self.__gpx.name

    @title.setter
    def title(self, value: str):
        """see getter."""
        if self.__gpx.name != value:
            self._load_full()
            self.__gpx.name = value
            self._dirty = 'title'

    @property
    def description(self) ->str:
        """str: The description.

        Returns:
            The description

        """
        if self.__gpx.description == Gpx.undefined_str:
            self._load_full()
            if self.__gpx.description == Gpx.undefined_str:
                self.__gpx.description = ''
        return self.__gpx.description

    @description.setter
    def description(self, value: str):
        """see getter."""
        if self.__gpx.description != value:
            self._load_full()
            self.__gpx.description = value
            self._dirty = 'description'

    @contextmanager
    def fenced(self, fences):
        """Suppress points in fences.

        While this context manager is running, suppressed points are
        not visible.

        """
        if not fences:
            yield
            return

        if self.__without_fences is not None:
            raise Exception('fenced() is already active')
        self.__without_fences = self.__gpx
        try:
            for track in self.__gpx.tracks:
                for segment in track.segments:
                    segment.points = [x for x in segment.points if fences.outside(x)]
            self.__gpx.waypoints = [x for x in self.__gpx.waypoints if fences.outside(x)]
            self._clear_similarity_cache()
            yield
        finally:
            self.__gpx = self.__without_fences
            self._clear_similarity_cache()
            self.__without_fences = None

    @contextmanager
    def _decouple(self):
        """Context manager: disable automic synchronization with the backend.

        In that state, automatic writes of changes into
        the backend are disabled, and if you access attributes which
        would normally trigger a full load from the backend, they will not.
        (The latter is used by __str__ and __repr__).

        """
        from_backend = self.__backend
        prev_value = from_backend._decoupled if from_backend is not None else None
        if from_backend is not None:
            from_backend._decoupled = True
        try:
            yield
        finally:
            if from_backend is not None:
                from_backend._decoupled = prev_value

    @property
    def __is_decoupled(self):
        """True if we are currently decoupled from the backend.

        In that state, changes to Track are not written to the backend and
        the track is not marked dirty.

        Returns:
            True if we are decoupled

        """
        if self.backend is None:
            return True
        return self.backend._decoupled

    @contextmanager
    def batch_changes(self):
        """Context manager: disable the direct update in the backend and saves the entire track when done.

        This may or may not make things faster.
        :class:`~gpxity.backends.directory.Directory` and
        :class:`~gpxity.backends.gpsies.GPSIES` profits from this,
        :class:`~gpxity.backends.mmt.MMT` maybe.

        """
        prev_batch_changes = self._batch_changes
        self._batch_changes = True
        try:
            yield
        finally:
            self._batch_changes = prev_batch_changes
            self._rewrite()

    @property
    def category(self) ->str:
        """str: What is this track doing? If we have no current value, return the default.

        The value is automatically translated between our internal value and
        the value used by the backend. This happens when reading from
        or writing to the backend. Here we return always the internal value.

        Returns:
            The current value or the default value (see :attr:`categories`)

        """
        if self.__gpx.category == Gpx.undefined_str:
            self._load_full()
        if self.__gpx.category == Gpx.undefined_str:
            return self.categories[0]
        if self.backend is None:
            return self.__gpx.category
        return self.backend.decode_category(self.__gpx.category)

    def __default_category(self) ->str:
        """The default for either an unsaved track or for the corresponding backend.

        Returns: The category

        """
        if self.backend is None:
            return self.categories[0]
        return self.backend.decode_category(self.backend.supported_categories[0])

    @category.setter
    def category(self, value: str):
        """see getter."""
        if value is None:
            value = self.__default_category()
        if value not in self.categories:
            raise Exception('Category {} is not known'.format(value))
        if value != self.__gpx.category:
            self._load_full()
            self.__gpx.category = value
            self.__encode_gpx()
            self._dirty = 'category'

    def _load_full(self):
        """Load the full track from source_backend if not yet loaded and if not decoupled.

        The backend may
         -  add values to self.__gpx
         -  replace self.__gpx, see gpx.setter

         The backend is allowed and expected to replace already known values. This may
         happen
         - if the backend has a mistake and returns different values in the list of the track
           and in the full downloaded track
         - if somebody else changed the track in the backend meanwhile

        Returns: True for success

        """
        if (self.backend is not None and self.id_in_backend and not self.__gpx.is_complete
                and not self.__is_decoupled and 'scan' in self.backend.supported):  # noqa
            self.backend._read_all_decoupled(self)

    def add_points(self, points) ->None:
        """Round and add points to last segment in the last track.

        If no track is allocated yet and points is not an empty list, allocates a track.

        Args:
            points (list(GPXTrackPoint): The points to be added

        """
        if points:
            self._round_points(points)
            self.gpx.add_points(points)
            self._dirty = 'gpx'

    def __decode_category(self, value) -> str:
        """Helper for _decode_keywords.

        Returns:
            A value out of Track.categories

        """
        return self.backend.decode_category(value) if self.backend is not None else value

    @staticmethod
    def _round_points(points):
        """Round points to 6 decimal digits because some backends may cut last digits.

        Gpsies truncates to 7 digits. The points are rounded in place!

        Args:
            points (list(GPXTrackPoint): The points to be rounded

        """
        for _ in points:
            _.longitude = round(_.longitude, 6)
            _.latitude = round(_.latitude, 6)

    def xml(self) ->str:
        """Produce exactly one line per trackpoint for easier editing (like removal of unwanted points).

        Returns: The xml string.

        """
        self._load_full()
        if self.__is_decoupled:
            # if Backend._write_all() calls this, everything is already encoded in __gpx
            self.__encode_gpx()   # TODO: should not be necessary
        return self.__gpx.xml()

    @property
    def public(self):
        """
        bool: Is this a private track (can only be seen by the account holder) or is it public?.

            Default value is False

        Returns:
            True if track is public, False if it is private

        """
        if self.__gpx.keywords == Gpx.undefined_str:
            self._load_full()
        if self.__gpx.public == Gpx.undefined_str:
            return False
        return self.__gpx.public

    @public.setter
    def public(self, value):
        """Store this flag as keyword 'public'."""
        if value not in (True, False):
            raise ValueError('public must be True or False, I got {}'.format(value))
        if value != self.__gpx.public:
            self._load_full()
            self.__gpx.public = value
            self.__gpx.encode()
            self._dirty = 'public'

    @property
    def gpx(self) ->Gpx:
        """
        Direct access to the Gpx object.

        If you use it to change its content
        * remember to call :meth:`rewrite` afterwards.
        * Since everything is stored in Gpx, all attributes like
        :attr:`description`, ;attr:`public` ... will change too.

        Returns:
            the Gpx object

        """
        self._load_full()
        return self.__gpx

    @gpx.setter
    def gpx(self, value):
        """Assign new gpx."""
        self.__gpx = value
        self._clear_similarity_cache()
        if self.__is_decoupled:
            self.__decode_gpx()
        else:
            self.__encode_gpx()
        self._round_points(self.points())

    @property
    def last_time(self) ->datetime.datetime:
        """The last time we received.

        Returns:
            The last time we received so far. If none, return None.

        """
        return self.gpx.last_time

    @property
    def keywords(self):
        """list(str): represent them as a sorted list - in GPX they are comma separated.

            Content is whatever you want.

            Because the GPX format does not have attributes for everything used by all backends,
            we encode some of the backend arguments in keywords.

            Example for mapmytracks: keywords = 'Status:public, Category:Cycling'.

            Gpxity expects keywords to be separated by ",". When writing them Gpxity
            uses ", " (with a space after the comma) as separator.

            However this is transparent for you. When parsing theGPX file, those are removed
            from keywords, and the are re-added in when exporting in :meth:`xml`. So
            :attr:`Track.keywords` will never show those special values.

            Some backends may change keywords. :class:`~gpxity.backends.mmt.MMT` converts the
            first character into upper case and will return it like that. Gpxity will not try to hide such
            problems. So if you save a track in :class:`~gpxity.backends.mmt.MMT`, its keywords
            will change. But they will not change if you copy from :class:`~gpxity.backends.mmt.MMT`
            to :class:`~gpxity.backends.directory.Directory` - so if you copy from DirectoryA
            to :class:`~gpxity.backends.mmt.MMT` to DirectoryB, the keywords in
            DirectoryA and DirectoryB will not be identical, for example "berlin" in DirectoryA but
            "Berlin" in DirectoryB.

        """
        return self.gpx.real_keywords

    @keywords.setter
    def keywords(self, values):
        """Replace all keywords.

        Args:
            Either single str with one or more keywords, separated by commas
                or an iterable of keywords. The new keywords. Must not have duplicates.

        """
        self.change_keywords(values, replace=True)

    @staticmethod
    def _check_keyword(keyword):
        """Must not be one of our internally used codes."""
        internal = (('Category', 'category'), ('Status', 'public'), ('Id', 'ids'))
        for internal_kw, attr in internal:
            if keyword.startswith(internal_kw + ':'):
                raise Exception('Do not use {} directly, use Track.{}'.format(internal_kw, attr))
        if ',' in keyword:
            raise Exception('No comma allowed within a keyword')

    def __prepare_keywords(self, values):
        """Common introductory code for change_keywords.

        The values may be preceded with a '-' which will be preserved in the result.__ids

        Args:
            values: Either single str with one or more keywords, separated by commas
                or an iterable of keywords.

        Returns:
            A set of legal keywords as expected by the backend.

        """
        if isinstance(values, str):
            lst = [x.strip() for x in values.split(',')]
        else:
            lst = list(values)
        pairs = list()
        for _ in lst:
            if _.startswith('-'):
                pairs.append((False, _[1:]))
            else:
                pairs.append((True, _))
        for _ in pairs:
            self._check_keyword(_[1])
        if self.backend is not None:
            pairs = [(x[0], self.backend._encode_keyword(x[1])) for x in pairs]
        adding = {x[1] for x in pairs if x[0]}
        removing = {x[1] for x in pairs if not x[0]}
        return adding, removing

    def change_keywords(self, values, replace=False, dry_run=False):
        """Change keywords.

        Duplicate keywords are silently ignored.
        A keyword may not contain a comma.
        Keywords with a preceding '-' are removed, the others are added.
        Raise an Exception if a keyword is both added and removed.

        Args:
            values: Either a single str with one or more keywords, separated by commas
                or an iterable of keywords
            replace: if True, replace current keywords with the new ones. Ignores keywords
                preceded with a '-'.
            dry_run: if True, only return the new keywords but do not make that change

        Returns:
            The new keywords

        """
        self._load_full()
        have = {*self.keywords}
        add, remove = self.__prepare_keywords(values)
        if add & remove:
            raise Exception('Cannot both add and remove keywords {}'.format(add & remove))

        if replace:
            remove = have - add
        else:
            add -= have
            remove &= have
        new = sorted((have | add) - remove)
        if not dry_run and self.keywords != new:
            self.__gpx.real_keywords = new
            with self.batch_changes():
                if remove:
                    self._dirty = 'remove_keywords{}{}'.format(BackendBase._dirty_separator, ', '.join(remove))
                if add:
                    self._dirty = 'add_keywords{}{}'.format(BackendBase._dirty_separator, ', '.join(add))
        assert new == self.keywords, (
            'change_keywords failed. Expected: {}, got: {}'.format(new, self.keywords))
        return new

    def speed(self) ->float:
        """Speed over the entire time in km/h or 0.0.

        Returns:
            The speed

        """
        return self.gpx.speed()

    def moving_speed(self) ->float:
        """Speed for time in motion in km/h.

        Returns:
            The moving speed

        """
        return self.gpx.moving_speed()

    def warnings(self):
        """Return a list of strings with easy to find problems."""
        result = list()
        if self.last_time:
            speed = self.speed()
            moving_speed = self.moving_speed()
            if speed > moving_speed:
                result.append('Speed {:.3f} must not be above Moving speed {:.3f}'.format(
                    speed, moving_speed))
            expected_speeds = {
                'Cycling - MTB': ((3, 50), (5, 50)),
                'Cycling - Road': ((3, 60), (10, 60)),
            }
            expected_speed = expected_speeds.get(self.category)
            if expected_speed:
                template = None
                if speed < expected_speed[0][0]:
                    template = 'Speed {speed:.3f} is very low'
                elif speed > expected_speed[0][1]:
                    template = 'Speed {speed:.3f} is very high'
                elif moving_speed < expected_speed[1][0]:
                    template = 'Moving speed {moving_speed:.3f} is very low'
                elif moving_speed > expected_speed[1][1]:
                    template = 'Moving speed {moving_speed:.3f} is very high'
                if template:
                    result.append(template.format(speed=speed, moving_speed=moving_speed))
        return result

    def __repr__(self) ->str:
        """The repr.

        Returns:
            the repr str

        """
        with self._decouple():
            # this should not automatically load the entire track
            parts = []
            parts.append('public' if self.public else 'private')
            if self.__gpx:
                parts.append(self.category)
                if self.keywords:
                    parts.append(','.join(self.keywords))
                if self.title:
                    parts.append(self.title)
                if self.first_time and self.last_time:
                    parts.append(repr_timespan(self.first_time, self.last_time))
                elif self.first_time:
                    parts.append(str(self.first_time))
                if self.distance:
                    parts.append('{:4.2f}km'.format(self.distance))
            return '{}({})'.format(str(self), ' '.join(parts))

    @staticmethod
    def identifier(backend, ident: str) ->str:
        """The full identifier for a track.

        Since we may want to do this without instantiating a track,
        this must be staticmethod or classmethod.

        str(track) uses this. However if a track has no id_in_backend,
        str(track will create one using title, time, id(track).

        Args:
            backend: May be :class:`~gpxity.backend.Backend` or :class:`~gpxity.accounts.Account`
            ident: id_in_backend

        Returns:
            the full identifier.

        """
        if isinstance(backend, BackendBase):
            account = backend.account
        else:
            account = backend
        if account is None:
            return 'unsaved: "{}"'.format(ident)
        if not ident:
            ident = 'no id_in_backend'
        return str(account) + ident

    def __str__(self) ->str:
        """The str.

        Returns:
            a unique full identifier

        """
        ident = self.id_in_backend
        if not ident:
            ident = '"{}" time={} id={}'.format(self.title or 'untitled', self.first_time, id(self))
        return self.identifier(self.backend, ident)

    def key(self, with_category: bool = True, with_last_time: bool = True, precision=None) ->str:
        """For speed optimized equality checks, not granted to be exact, but sufficiently safe IMHO.

        Args:
            with_category: If False, do not use self.category. Needed for comparing
                tracks for equality like in unittests because values can change
                and information can get lost while copying between different
                backends
            with_last_time: If False, do not use self.last_time.
            precision: For latitude/longitude. After comma digits. Default is as defined by backend or 6.

        Returns:
            a string with selected attributes in printable form.

        """
        self._load_full()
        return 'title:{} description:{} keywords:{} category:{}: public:{} last_time:{} angle:{} points:{}'.format(
            self.title, self.description,
            ','.join(self.keywords).lower(), self.category if with_category else '',
            self.public, self.last_time if with_last_time else '',
            self.angle(precision), self.gpx.get_track_points_no())

    def __eq__(self, other) ->bool:
        """equal.

        Returns:
            result

        """
        if self is other:
            return True
        return self.key() == other.key()

    def __lt__(self, other) ->bool:
        """less than.

        Returns:
            result

        """
        return self.key() < other.key()

    def angle(self, precision=None) ->float:
        """For me, the earth is flat.

        Args:
            precision: After comma digits. Default is as defined by backend or 6.

        Returns:
            the angle in degrees 0..360 between start and end.
            If we have no track, return 0

        """
        return self.gpx.angle(precision)

    def segments(self):
        """
        A generator over all segments.

        Yields:
            GPXTrackSegment: all segments in all tracks

        """
        for _ in self.gpx.segments():
            yield _

    def points(self):
        """
        A generator over all points.

        Yields:
            GPXTrackPoint: all points in all tracks and segments

        """
        for _ in self.gpx.points():
            yield _

    def point_list(self):
        """A flat list with all points.

        Returns:
            The list

        """
        return sum((x.points for x in self.segments()), [])

    def last_point(self):
        """Return the last point of the track. None if none."""
        # TODO: unittest for track without __gpx or without points
        return self.gpx.last_point()

    def adjust_time(self, delta):
        """Add a timedelta to all times."""
        self.gpx.adjust_time(delta)
        self._dirty = 'gpx'

    def points_hash(self) -> float:
        """A hash that is hopefully different for every possible track.

        It is built using the combination of all points.

        Returns:
            The hash

        """
        return self.gpx.points_hash()

    def points_equal(self, other, digits=4) ->bool:
        """
        Compare points for same position.

        Args:
            digits: Number of after comma digits to compare

        Returns:
            True if both tracks have identical points.

        All points of all tracks and segments are combined. Elevations are ignored.

        """
        return self.gpx.points_equal(other.gpx, digits)

    def index(self, other, digits=4):
        """Check if this track contains other track.gpx.

        This only works if all values for latitude and longitude are
        nearly identical.

        Useful if one of the tracks had geofencing applied.

        Args:
            digits: How many after point digits are used

        Returns:
            None or the starting index for other.points in self.points

        """
        return self.gpx.index(other.gpx, digits)

    @staticmethod
    def overlapping_times(tracks):
        """Find tracks with overlapping times.

        Yields:
            groups of tracks with overlapping times. Sorted by time.

        This may be very slow for many long tracks.

        """
        previous = None
        group = list()  # Track is  mutable, so a set is no possible
        for current in sorted(tracks, key=lambda x: x.first_time):
            if previous and current.first_time <= previous.last_time:
                if previous not in group:
                    group.append(previous)
                group.append(current)
            else:
                if group:
                    yield sorted(group, key=lambda x: x.first_time)
                    group = list()
            previous = current
        if group:
            yield sorted(group, key=lambda x: x.first_time)

    def _has_default_title(self) ->bool:
        """Try to check if track has the default title given by a backend.

        Returns:

            True if so"""
        # the title of MMT might have been copied into another backend:
        if not self.title:
            return True
        if self.title == '{} track'.format(self.category):
            return True
        if all(x in '0123456789 :-_' for x in self.title):
            return True
        return False

    def __merge_metadata(self, other, dry_run):
        """Merge metadata from other.

        Used only by self.merge.

        Returns:
            a list of verbosity messages

        """
        msg = list()
        if not other._has_default_title() and self._has_default_title():
            msg.append('Title: {} -> {}'.format(self.title, other.title))
            if not dry_run:
                self.title = other.title
        if other.description and other.description != self.description:
            msg.append('Additional description: {}'.format(
                other.description))
            if not dry_run:
                self.description += '\n'
                self.description += other.description
        if other.public and not self.public:
            msg.append('Visibility: private -> public')
            if not dry_run:
                self.public = True
        if other.category != self.category:
            msg.append('Category: {}={} wins over {}={}'.format(
                other, other.category, self, self.category))
        kw_src = set(other.keywords)
        kw_dst = set(self.keywords)
        if kw_src - kw_dst:
            msg.append('New keywords: {}'.format(','.join(kw_src - kw_dst)))
            if not dry_run:
                self.keywords = kw_src | kw_dst
        # ids are diffent. keywords are sorted, ids is fifo
        new_ids = self.__clean_ids(self.ids + other.ids)
        if new_ids != self.ids:
            msg.append('New Ids: {}'.format(','.join(set(new_ids) - set(self.ids))))
            if not dry_run:
                self.ids = new_ids
        return msg

    def can_merge(self, other, partial_tracks: bool = False):
        """Check if self and other are mergeable.

        args:
            other: The other Track
            partial_tracks: If True, they are mergeable if one of them contains
                the other one.

        Returns: (int, str)
            int is either None or the starting index of the shorter track in the longer track
            str is either None or a string explaing why this is not mergeable

        """
        if str(self) == str(other):
            return None, 'Cannot merge identical tracks {}'.format(self)

        if (other.gpx.get_track_points_no()) == 0 and other.gpx.waypoints:
            # mergable
            return 0, None

        if partial_tracks:
            other_in_self = self.index(other)
            if other_in_self is not None:
                return other_in_self, None
            self_in_other = other.index(self)
            if self_in_other is not None:
                return self_in_other, None
        elif self.points_equal(other):
            return 0, None
        return None, (
            'Cannot merge {} with {} points into {} with {} points'.format(
                other, other.gpx.get_track_points_no(),
                self, self.gpx.get_track_points_no()))

    def __merge_waypoints(self, other, dry_run) ->list:
        """Merge waypoints from other with differing positions.

        Returns: list(str)
            Messages about what has been done.

        """
        # TODO: unittest
        def have(wpt):
            """True if we already have this waypoint."""
            return (wpt.latitude, wpt.longitude) in [(x.latitude, x.longitude) for x in new_wpts]
        merged_count = 0
        new_wpts = self.gpx.waypoints[:]
        for other_wpt in other.gpx.waypoints:
            if not have(other_wpt):
                new_wpts.append(other_wpt)
                if not dry_run:
                    self.gpx.waypoints.append(other_wpt)
                    self.rewrite()
                merged_count += 1
        if merged_count > 0:
            return ['{} got {} waypoints from {}'.format(self, merged_count, other)]
        return []

    def __merge_tracks(self, other, dry_run, shorter_at) ->list:
        """Merge tracks from other.

        Returns: list(str)
            Messages about what has been done.

        """
        msg = []
        if other.gpx.get_track_points_no() > self.gpx.get_track_points_no():
            if not dry_run:
                self.gpx.tracks = deepcopy(other.gpx.tracks)
                self.rewrite()
            msg.append('{} got entire gpx.tracks from {}'.format(self, other))
        changed_point_times = 0
        self_points = self.point_list()[shorter_at:]
        for self_point, other_point in zip(self_points, other.points()):
            # TODO: unittest with shorter track
            if not self_point.time:
                if not dry_run:
                    self_point.time = other_point.time
                changed_point_times += 1
        if changed_point_times:
            if not dry_run:
                self.rewrite()
            msg.append('Copied times for {} out of {} points'.format(
                changed_point_times, self.gpx.get_track_points_no()))
        return msg

    def merge(  # noqa pylint: disable=unused-argument
            self, other, remove: bool = False, dry_run: bool = False, copy: bool = False,
            partial_tracks: bool = False) ->list:
        """Merge other track into this one.

        Either the track points must be identical or the other track
        may only contain waypoints.

        If merging is not possible, raise Track.CannotMerge.

        If either is public, the result is public.
        If self.title seems like a default and other.title does not, use other.title
        Combine description and keywords.
        Merge waypoints as defined by _merge_waypoints().

        Args:
            other (:class:`~gpxity.track.Track`): The track to be merged
            remove: After merging succeeded, remove other
            dry_run: if True, do not really apply the merge
            copy: This argument is ignored. It is only here to give :meth:`Track.merge() <gpxity.track.Track.merge>`
                and :meth:`Backend.merge() <gpxity.backend.Backend.merge>` the same interface.
            partial_tracks: merges other track
                if either track is part of the other one

        Returns: list(str)
            Messages about what has been done.

        """
        assert isinstance(other, Track)
        msg = []
        shorter_at, _ = self.can_merge(other, partial_tracks)
        if shorter_at is None:
            raise Track.CannotMerge(_)
        with self.batch_changes():
            msg.extend(self.__merge_tracks(other, dry_run, shorter_at))
            msg.extend(self.__merge_waypoints(other, dry_run))
            msg.extend(self.__merge_metadata(other, dry_run))
        if msg:
            msg = ['     ' + x for x in msg]
            msg.insert(0, 'merge{} {!r}'.format(' and remove' if remove else '', other))
            msg.insert(1, '{}  into {!r}'.format(' ' * len(' and remove') if remove else '', self))
        if remove:
            if len(msg) <= 2:
                msg.append(
                    'remove duplicate {!r}: It was identical with {!r}'.format(other, self))
            if not dry_run:
                other.remove()
        return msg

    def fix_orux(self):
        """Older Oruxmaps switched the day back by one day after exactly 24 hours.

        Try fixing that.

        Please backup your track before doing this.

        Returns: A list with messages. Currently nothing.

        """
        if self.gpx.fix_orux():
            self._dirty = 'gpx'
        return []

    def split_at_stops(self, minutes):
        """Split where things happen.

        Whenever the time jumps back or more than 30
        minutes into the future, split the segment at that point.

        Returns:
            A list of message strings, usable for verbose output.

        """
        if self.gpx.fix_jumps(minutes=minutes):
            self._dirty = 'gpx'

    def __similarity_to(self, other):
        """Return a float 0..1: 1 is identity."""

        def simple(track):
            """Simplified track"""
            return [(round(x.latitude, 3), round(x.longitude, 3))
                    for x in simplify_polyline(list(track.points()), max_distance=50)]

        if id(other) not in self._similarity_others:
            simple1 = simple(self)
            simple2 = simple(other)
            max_len = max([len(simple1), len(simple2)])
            similar_length = 1.0 - abs(len(simple1) - len(simple2)) / max_len
            set1 = set(simple1)
            set2 = set(simple2)
            min_len = min([len(set1), len(set2)])
            similar_points = len(set1 & set2) / min_len
            result = similar_length * similar_points
            self._similarity_others[id(other)] = other
            self._similarities[id(other)] = result
            other._similarity_others[id(self)] = self
            other._similarities[id(self)] = result
        return self._similarities[id(other)]

    def similarity(self, others):
        """Return a float 0..1: 1 is identity.

        The highest value for others is returned."""

        if isinstance(others, Track):
            others = [others]
        return max(self.__similarity_to(x) for x in others)

    def time_offset(self, other):
        """If time and last_time have the same offset between both tracks, return that time difference.
        Otherwise return None."""
        return self.gpx.time_offset(other.gpx)

    @property
    def ids(self):
        """Return ids for all backends where this track has already been.

        You can modify it but your changes will only be saved if and when
        the entire track is saved.

        They are sorted by ascending age.
        Only the 5 youngest are kept.
        If the same id_in_backend appears in more than one directory, keep only the youngest.

        Returns: list( (str))
            a list of track ids

        """
        if self.__gpx.keywords == Gpx.undefined_str:
            self._load_full()
        return self.__clean_ids(self.__gpx.ids)

    @ids.setter
    def ids(self, value):
        """Setter for ids."""
        cleaned = self.__clean_ids(value)
        if cleaned != self.__gpx.ids:
            self._load_full()
            self.__gpx.ids = cleaned
            self.__gpx.encode()

    @staticmethod
    def __clean_ids(original):
        """Remove redundancies and old ids.append.

        1. if the same id_in_backend is in several directories, keep only the first one
        2. keep only 5

        Returns: The cleaned list if ids

        """
        result = list()
        seen_url = set()
        seen_id = set()
        for orig_id in original:
            if orig_id in seen_id:
                continue
            seen_id.add(orig_id)
            try:
                acc, _ = BackendBase.parse_objectname(orig_id)
            except KeyError:
                continue
            if acc.backend == 'Directory':
                if acc.url not in seen_url:
                    seen_url.add(acc.url)
                    result.append(orig_id)
            else:
                result.append(orig_id)
        result = result[:5]
        if result != original:
            logging.debug('ids: %s -> %s', original, result)
        return result

    def split_segments(self):
        """Create separate tracks for every track/segment."""
        backend = self.backend
        clone = self.clone()
        self.remove()
        try:
            for segment in clone.segments():
                track = self.clone()
                gpx_track = GPXTrack()
                gpx_track.segments.append(segment)
                track.gpx.tracks = [gpx_track]
                backend.add(track)
        except BaseException as exc:
            logging.error('split:%s', exc)
            backend.add(clone)

    def locate_point(self, track=0, segment=0, point=0) ->str:
        """Determine name of place for point.

        Saves that in point.name for caching. If backend.account.country is given,
        add the country name only if it is different.

        Args:
            track, segment, point: Indices into the respective arrays
            default_country: suppress this one in the result

        Returns: A string

        """
        country = self.backend.account.country if self.backend is not None else None
        result, located = self.gpx.locate_point(track, segment, point, default_country=country)
        if located:
            if self.backend is not None:
                self.rewrite()
        return result

    def add_locations(self, segments=False):
        """Call locate_point for the first point.

        Args: segments: Also do that for the first point of each segment.

        """
        self.locate_point()
        if segments:
            for track_idx, gpx_track in enumerate(self.gpx.tracks):
                for seg_idx, _ in enumerate(gpx_track.segments):
                    self.locate_point(track=track_idx, segment=seg_idx)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) 2019 Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.
# The source in this file is inspired by and partially identical with paramiko.config


"""Configuration file for accounts in Backends."""

import os
import re
import copy
import tempfile

from gpxpy.geo import Location

__all__ = ['Account', 'DirectoryAccount']


class Fences:  # pylint: disable=too-few-public-methods

    """
    Defines circles.

    Args:
        config_str: The string from the accounts file
    Attributes:
        center (GPXTrackPoint): The center
        radius (meter): The radius in meters

    """

    def __init__(self, config_str: str):
        """init."""
        self.circles = list()
        if config_str is not None:
            for fence in config_str.split(' '):
                parts = fence.split('/')
                if len(parts) != 3:
                    raise ValueError('fence needs 3 parts: {}'.format(fence))
                try:
                    parts = [x.strip() for x in parts]
                    center = Location(float(parts[0]), float(parts[1]))
                    radius = float(parts[2])
                except Exception:
                    raise ValueError('Fence definition is wrong: {}'.format(fence))
                circle = (center, radius)
                self.circles.append(circle)

    def outside(self, point) ->bool:
        """Determine if point is outside of all fences.

        Returns: True or False.

        """
        return all(point.distance_2d(x[0]) > x[1] for x in self.circles)


class Accounts:

    """Representation of config information as stored in the format used by Gpxity.

    Queries can be made via `lookup`. The keyword  :literal:`Account` only allows one name.

    Keywords are case insensitive, arguments are not.

    Example for an entry in the accounts file:

    ::

        Account wp
            Backend WPTrackserver
            Username wordpress_username
            Url localhost
            Mysql wordpress_7@wordpress_7
            Password xxxx
            Fences 53.7505,10.7445/750

    """

    # pylint: disable=too-few-public-methods

    __SETTINGS_REGEX = re.compile(r'(\w+)(?:\s*=\s*|\s+)(.+)')

    __account_files = dict()

    @classmethod
    def __parse(cls, path):
        """Parse an accounts file."""
        if path not in cls.__account_files:
            if not os.path.exists(path):
                return
            cls.__account_files[path] = cls.__parse_accounts(path)

    @classmethod
    def __parse_accounts(cls, filename):
        """Parse all accounts from filename.

        Returns: dict with all accounts.filename

        """
        result = dict()
        with open(filename) as file_obj:
            for _ in cls.__yield_accounts(file_obj):
                result[_['name']] = _
        return result

    @staticmethod
    def __strip_whitespace(file_obj):
        """Filter out comments, strip lines."""
        for line in file_obj:
            line = line.strip()
            if line and not line.startswith('#'):
                yield line

    @classmethod
    def __yield_matches(cls, file_obj):
        """Yield usable lines."""
        for line in cls.__strip_whitespace(file_obj):
            match = re.match(cls.__SETTINGS_REGEX, line)
            if not match:
                raise Exception('Unparsable line {}'.format(line))
            yield match

    @classmethod
    def __yield_accounts(cls, file_obj):
        """Generate all accounts."""
        account = {'name': 'global'}
        for match in cls.__yield_matches(file_obj):
            key = match.group(1).lower()
            value = match.group(2)

            if key == 'account':
                if account is not None:
                    yield account
                account = {'name': value.lower()}
                continue

            if value.startswith('"') and value.endswith('"'):
                value = value[1:-1]

            if key not in account:
                account[key] = value

        if account is not None:
            yield account

    @classmethod
    def lookup(cls, filename: str, wanted_account: str):
        """
        Build an :class:`~gpxity.accounts.Account`.

        Args:
            filename: The name of the accounts file
            wanted_account: The name to look for in the accounts file

        Returns: :class:`~gpxity.accounts.Account`

        """
        cls.__parse(filename)
        return copy.deepcopy(cls.__account_files[filename][wanted_account.lower()])


class Account:

    """As parsed from the accounts file.

    Attributes can be referenced as account.xxxx where xxx is an arbitrary
    value in the account definition from the accounts file.

    Args:
        name: The name of the account. Must exist in the accounts file.
        filename: Name of the accounts file. Default is Account.path
        kwargs: Additional parameters added to the account. They have precedence.

    If both name and file are None, only :literal:`**kwargs` are used.

    Attributes:
        path: Default value for the accounts file
        name: The name of the account
        config: A dict with all config values
        backend: The name of the backend class
        fences: The backend will never write points within fences.
            You can define any number of fences separated by spaces. Every fence is a circle.
            It has the form Lat/Long/meter.
            Lat and Long are the center position in decimal degrees, meter is the radius.

    """

    path = '~/.config/Gpxity/accounts'

    def __init__(self, name=None, filename=None, **kwargs):
        """Create an Account."""
        if name is None:
            self.config = dict()
            for key, value in kwargs.items():
                self.config[key.lower()] = value
            self.name = self.url or '.'
            if not self.backend:
                self.config['backend'] = 'Directory'
            self._resolve_fences()
            return
        self.name = name
        path = os.path.expanduser(filename or Account.path)
        lookup_name = name.split(':')[0]
        self.config = Accounts.lookup(path, lookup_name)
        if self.backend is None:
            raise Exception('Account({}, {}, {}) defines no Backend'.format(name, filename, kwargs))
        for key, value in kwargs.items():
            self.config[key.lower()] = value
        self._resolve_fences()

    def _resolve_fences(self):
        """create self.fences as a Fences instance."""
        if 'fences' in self.config:
            _ = Fences(self.config['fences'])
            del self.config['fences']
            self.fences = _
        else:
            self.fences = Fences(None)

    def __getattr__(self, key):
        """Only called if key is not an existing attribute.

        Returns: The value or None

        """
        try:
            config = object.__getattribute__(self, 'config')
        except AttributeError:
            return None
        return config.get(key.lower())

    def __repr__(self):
        """For debugging output.

        Returns: the str

        """
        result = 'Account({}): backend={}'.format(self.account, self.backend)
        if 'url' in self.config:
            result += ' url={}'.format(self.url)
        if 'username' in self.config:
            result += ' username={}'.format(self.username)
        return result + ')'

    def __str__(self):
        """The account in a parseable form.

        Returns: The string

        """
        return self.name + ':'


class DirectoryAccount(Account):

    """This will not use an acocunts file but the optional file :literal:`.config`.

    Args:
        url: The name of the directory. If it does not exist, create it.
            "" will translate into ".".
            A trailing "/" will raise an Exception.
            None will create a temporary directory.
        kwargs: Additional parameters added to the account. They have precedence.

    Attributes:
        path: Default value for the accounts file
        name: The name of the account
        config: A dict with all config values
        backend: The name of the backend class
        is_temporary: True for temporary directories.
        fences: The backend will never write points within fences.
            You can define any number of fences separated by spaces. Every fence is a circle.
            It has the form Lat/Long/meter.
            Lat and Long are the center position in decimal degrees, meter is the radius.
        prefix (str):  Class attribute, may be changed. The default prefix for
            temporary directories. Default value is :literal:`gpxity.`

    """

    path = None
    prefix = 'gpxity.'

    def __init__(self, url=None, **kwargs):  # pylint: disable=super-init-not-called
        """Create an Account."""
        self.is_temporary = url is None
        if self.is_temporary:
            url = tempfile.mkdtemp(prefix=self.__class__.prefix)
        if url == '':
            url = '.'
        if url == '/':
            raise Exception('Directory / is not allowed')
        if url.endswith('/') and url != '/':
            raise Exception('DirectoryAccount: url {} must not end with /'.format(url))
        self.config = dict()
        if not os.path.exists(url):
            os.makedirs(url)
        else:
            config_name = os.path.join(url, '.gpxity_config')
            if os.path.exists(config_name):
                self.config = Accounts.lookup(config_name, 'global')
        self.config['backend'] = 'Directory'
        self.config['url'] = url
        for key, value in kwargs.items():
            self.config[key.lower()] = value
        self.name = url
        self._resolve_fences()

    def __repr__(self):
        """For debugging output.

        Returns: the str

        """
        return 'DirectoryAccount({})'.format(self.name)

    def __str__(self):
        """The account in a parseable form.

        Returns: The string

        """
        if self.name == '.':
            return ''
        if self.name == '/':
            return '/'
        return self.name + '/'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) 2019 Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""This implements :class:`gpxity.mailer.Mailer`: a mailing backend. It can only write."""

# pylint: disable=protected-access

import datetime
from threading import Timer
import smtplib
import socket
import logging
from email.message import EmailMessage

from .. import Backend

__all__ = ['Mailer']


class MailQueue:

    """Holds all data representing a sent mail."""

    # pylint: disable=too-few-public-methods

    counter = 0

    def __init__(self, mailer):
        """See class docstring."""
        self.mailer = mailer
        self.disabled = False
        self.tracks = dict()
        self.last_sent_time = datetime.datetime.now() - datetime.timedelta(days=5)

    def append(self, track):
        """Append a track to the mailing queue."""
        self.tracks[track.id_in_backend] = track.clone()
        if hasattr(track, 'mail_subject'):
            self.tracks[track.id_in_backend].mail_subject = track.mail_subject

    def subject(self, track=None) ->str:
        """Build the mail subject.

        Args:
            track: If given, use only track. Otherwise use self.tracks

        Returns:
            The subject

        """
        if len(self.tracks) == 1:
            for _ in self.tracks.values():
                track = _
        if track is not None:
            subject = track.mail_subject if hasattr(track, 'mail_subject') else track.title
            return self.mailer.subject_template.format(
                title=subject, distance='{:8.3f}km'.format(track.distance))
        return '{} tracks'.format(len(self.tracks))

    def content(self) ->str:
        """The content of the mail message.

        Returns: list
            The content, a list with lines

        """
        is_single = len(self.tracks) == 1
        result = list()
        for key, track in self.tracks.items():
            if not key.endswith('.gpx'):
                key += '.gpx'
            if is_single:
                result.append(track.description)
                result.append('See the attached GPX file {}'.format(key))
            else:
                result.append(self.subject(track))
                indent = '   '
                result.append('{}See the attached GPX file {}'.format(indent, key))
                result.append('{}{}'.format(indent, track.description))
            result.append('')
            result.append('')
        return result

    def send(self):
        """Actually send the mail if there are any points."""
        if not self.tracks:
            return
        if self.disabled:
            self.tracks = dict()
            return
        account = self.mailer.account
        mail = EmailMessage()
        mail['subject'] = self.subject()
        mail['from'] = account.mailfrom or 'gpxity'
        mail['to'] = account.url.split()
        mail.set_content('\n'.join(self.content()))

        for key, track in self.tracks.items():
            if not key.endswith('.gpx'):
                key += '.gpx'
            mail.add_attachment(track.xml(), filename=key)
        host = account.smtp or 'localhost'
        port = int(account.port or '25')
        timeout = self.mailer.timeout
        if isinstance(timeout, (tuple, list)):
            timeout = timeout[0]
        try:
            with smtplib.SMTP(  # noqa
                    host,
                    port=port,
                    timeout=timeout) as smtp_server:
                smtp_server.send_message(mail)
        except socket.timeout:
            logging.error('Mailer: Disabled because the smtp server %s:%d did not answer within %d seconds ',
                          host, port, self.mailer.timeout)
            self.disabled = True
        except smtplib.SMTPRecipientsRefused as exc:
            logging.error('Mailer: Disabled because some Recipients are refused: %s', exc.recipients)
            self.disabled = True
            raise
        self.last_sent_time = datetime.datetime.now()
        self.mailer.history.append('to {}: {}'.format(mail['to'], mail['subject']))
        self.tracks = dict()

    def __repr__(self):
        """Return repr."""
        return 'MailQueue({} to {}'.format(', '.join(str(x) for x in self.tracks.values()), self.mailer.url)  # noqa


class Mailer(Backend):  # pylint: disable=abstract-method

    """Mailing backend. Write-only.

    Attributes:
        subject_template: This builds the mail subject. {title} and {distance} will
            be replaced by their respective values. Other placeholders are not yet defined.
        outstanding_tracks: Do not change this dict. Key is track.id_in_backend, value is
            a clone of the track. This freezes the current title in the clone.
        url: Holds the address of the recipient.
        account.mailfrom: The name of the mail sender. Default "gpxity".
        account.port: The port of the smtp server to talk to. Default 25
        account.smtp: The name of the smtp server. Default "localhost".
        account.interval (str): seconds. Mails are not sent more often. Default is None. If None, always send when
            Mailer.flush() is called. This is used for bundling several writes into one single mail:
            gpxdo merge --copy will send all tracks with one single mail.
            Lifetracking uses this to send mails with the current track only every X seconds, the
            mail will only contain the latest version of the track.

    """

    id_count = 0

    test_is_expensive = False

    def __init__(self, account):
        """See class docstring."""
        super(Mailer, self).__init__(account)
        self.history = list()
        self.subject_template = '{title} {distance}'
        self.timer = None
        self.queue = MailQueue(self)

    def _new_ident(self, _) ->str:
        """Build a unique id for track.

        Returns:
            A new unique id.

        """
        self.id_count += 1
        return str(self.id_count)

    def _write_all(self, track) ->str:
        """Mail the track.

        Returns:
            track.id_in_backend

        """
        if track.id_in_backend is None:
            new_ident = self._new_ident(track)
            with track._decouple():
                track.id_in_backend = new_ident
        self.queue.append(track)
        if self.account.interval is not None:
            seconds = int(self.account.interval)
            if self.queue.last_sent_time + datetime.timedelta(seconds=seconds) < datetime.datetime.now():
                self.queue.send()
            else:
                self._start_timer()
        return track.id_in_backend

    def detach(self):
        """Mail the rest."""
        if self.timer:
            self.timer.cancel()
        self.queue.send()

    def flush(self):
        """Now is the time to write."""
        # this dict reduces all tracks to just one instance
        self.queue.send()
        if self.timer:
            self.timer.cancel()
            self.timer = None

    def _start_timer(self, interval=None):
        """Start the flush timer."""
        if self.timer is None:
            if interval is None:
                interval = int(self.account.interval)
            self.timer = Timer(interval, self.flush)
            self.timer.start()

    def _lifetrack_start(self, track, points) ->str:
        """flush.

        Returns: The new id_in_backend.

        """
        track.mail_subject = 'Lifetracking starts: {}'.format(track.title)
        new_ident = self._write_all(track)
        self._append(track)
        assert self._has_item(new_ident), '{} not in {}'.format(new_ident, self)
        self.flush()
        return new_ident

    def _lifetrack_update(self, track, points):
        """flush."""
        track.mail_subject = 'Lifetracking continues: {}'.format(track.title)
        self._write_all(track)

    def _lifetrack_end(self, track):
        """flush."""
        track.mail_subject = 'Lifetracking ends: {}'.format(track.title)
        self._write_all(track)
        self.flush()

    def __str__(self) ->str:
        """A unique identifier.

        Returns:
            the unique identifier

        """
        return 'mailto:{}'.format(self.url)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) 2019 Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""This module defines :class:`~gpxity.locate.Locate`."""

import logging

import geocoder
from gpxpy.gpx import GPXTrackPoint

__all__ = ['Locate']


class Locate:

    """Locates tracks using https://github.com/DenisCarriere/geocoder#overview .

    Args:
        places: A list of places that the tracks should pass
        tracks: The tracks to be searched

    Attributes:
        locations: The list of found places. For each given value in arg **places**,
            locations holds one result, even if the provider returns more than one.
        distances: A list with a tuple for every track. The tuple holds the track
            and a list of distances between that track and places.

    """

    # pylint: disable=too-few-public-methods

    def __init__(self, places, tracks):
        """See class docstring."""
        self.places = places
        self.tracks = tracks
        self.locations = list()
        for place in places:
            _ = geocoder.get(place, provider='osm')
            if not _:
                raise Exception('Place not found: {}'.format(place))
            self.locations.append(_[0])
        logging.info('using locations:')
        for _ in self.locations:
            logging.info('  %s', _.address)
        self.distances = list()
        gpx_points = [GPXTrackPoint(latitude=x.lat, longitude=x.lng) for x in self.locations]
        for track in tracks:
            self.distances.append(
                (track, [track.gpx.get_nearest_location(x).location.distance_2d(x) for x in gpx_points]))
        self.distances.sort(key=lambda x: sum(x[1]))

    def found(self, max_away: float = 1e10):
        """The list of tracks sorted by affinity to the given places.

        Args:
            max_away: The maximum distance in kilometers

        Returns: list(tracks)

        """
        return [x[0] for x in self.distances if sum(x[1]) < max_away * 1000]
